import asyncio
import re
import typing
from asyncio import StreamWriter, StreamReader
from textwrap import wrap

from .shared import get_bot
from .config import get_nick
from .enums import Event
from .events import trigger_event
from .ratelimit import privmsg_ratelimit, whisper_ratelimit

if typing.TYPE_CHECKING:
    from .bots import BaseBot

PRIVMSG_MAX_LINE_LENGTH = 450
WHISPER_MAX_LINE_LENGTH = 438
PRIVMSG_FORMAT = 'PRIVMSG #{channel} :{line}'

class Irc:
    def __init__(self, reader, writer):
        self.reader: StreamReader = reader
        self.writer: StreamWriter = writer
        self.bot: 'BaseBot' = get_bot()

    def send(self, msg):
        """
        sends a raw message with no modifications, this function is not ratelimited!

        do not call this function to send channel messages or whisper,
        this function is not ratelimit and intended to internal use from 'send_privmsg' and 'send_whisper'
        only use this function if you need to
        """
        self.writer.write(f'{msg}\r\n'.encode())

    def send_all(self, *msgs):
        """
        sends all messages separately with no modifications, this function is not ratelimited!

        do not call this function to send channel messages or whisper,
        this function is not ratelimit and intended to internal use from 'send_privmsg' and 'send_whisper'
        only use this function if you need to
        """
        for msg in msgs:
            self.send(msg)

    async def send_all_interval(self, *msgs, delay=.2):
        """
        sends all messages separately with no modifications, this function is not ratelimited!
        delays messages with the delay passed in from the `delay` argument

        do not call this function to send channel messages or whisper,
        this function is not ratelimit and intended to internal use from 'send_privmsg' and 'send_whisper'
        only use this function if you need to
        """
        for msg in msgs:
            self.send(msg)
            await asyncio.sleep(delay)

    async def send_privmsg(self, channel: str, msg: str):
        """sends a message to a channel"""
        # import it locally to avoid circular import
        from .channel import channels, DummyChannel
        from .modloader import trigger_mod_event

        channel = channel.lower()
        chan = channels.get(channel) or DummyChannel(channel)
        for line in _wrap_message(msg):
            await privmsg_ratelimit(chan)
            self.send(PRIVMSG_FORMAT.format(channel=channel, line=line))

        # exclude calls from send_whisper being sent to the bots on_privmsg_received event
        if not msg.startswith('/w'):
            if self.bot:
                await self.bot.on_privmsg_sent(msg, channel, get_nick())
            await trigger_mod_event(Event.on_privmsg_sent, msg, channel, get_nick(), channel=channel)
            await trigger_event(Event.on_privmsg_sent, msg, channel, get_nick())

    async def send_whisper(self, user: str, msg: str):
        """sends a whisper to a user"""
        from .modloader import trigger_mod_event

        user = user.lower()
        for line in _wrap_message(f'/w {user} {msg}'):
            await whisper_ratelimit()
            self.send(PRIVMSG_FORMAT.format(channel=user, line=line))
            # this sleep is necessary to make sure all whispers get sent
            # without it, consecutive whispers get dropped occasionally
            # if i find a better fix, will do it instead, but until then, this works
            await asyncio.sleep(.6)

        if self.bot:
            await self.bot.on_whisper_sent(msg, user, get_nick())
        await trigger_mod_event(Event.on_whisper_sent, msg, user, get_nick())
        await trigger_event(Event.on_whisper_sent, msg, user, get_nick())

    async def get_next_message(self):
        return (await self.reader.readline()).decode('utf8').strip()

    def send_pong(self):
        self.send('PONG :tmi.twitch.tv')


def _wrap_message(msg):
    m = re.match(r'/w (?P<user>[\w\d_]+)', msg)
    if m:
        whisper_target = m['user']
    else:
        whisper_target = None
    whisper_prefix = f'/w {whisper_target}'

    for line in wrap(msg, width=PRIVMSG_MAX_LINE_LENGTH if not msg.startswith('/w') else WHISPER_MAX_LINE_LENGTH):
        if whisper_target and not line.startswith(whisper_prefix):
            line = f'{whisper_prefix} {line}'
        yield line



from sqlalchemy import create_engine, orm
from sqlalchemy.ext.declarative import declarative_base

from ..config import mysql_cfg

__all__ = ('Base', 'engine', 'session', 'DB_FILENAME', 'init_tables')

if mysql_cfg.enabled:
    try:
        import mysql.connector
    except (ImportError, ModuleNotFoundError):
        print(
            'Could not find mysql connection library, please install it via pip:\n\tpip install --upgrade --user mysql-connector-python')
        input('\npress enter to exit...')
        exit(1)

Base = declarative_base()
DB_FILENAME = 'database.sqlite'
engine = create_engine(f'sqlite:///{DB_FILENAME}'
                       if not mysql_cfg.enabled else
                       f'mysql+mysqlconnector://{mysql_cfg.username}:{mysql_cfg.password}@{mysql_cfg.address}:{mysql_cfg.port}/{mysql_cfg.database}')
Session = orm.sessionmaker(bind=engine)
session = orm.scoped_session(Session)


def init_tables():
    Base.metadata.create_all(engine)



from datetime import datetime, timedelta
from secrets import randbelow
from typing import Dict

from twitchbot import (
    Arena,
    ARENA_DEFAULT_ENTRY_FEE,
    Command,
    Message,
    set_currency_name,
    get_currency_name,
    set_balance,
    get_balance,
    session,
    get_balance_from_msg,
    add_balance,
    cfg,
    InvalidArgumentsError,
    duel_expired,
    add_duel,
    accept_duel,
    subtract_balance,
    get_duel,
    add_balance_to_all,
    Balance,
    subtract_balance_from_all,
    get_nick)

PREFIX = cfg.prefix
MANAGE_CURRENCY_PERMISSION = 'manage_currency'


@Command('setcurrencyname', permission=MANAGE_CURRENCY_PERMISSION, syntax='<new_name>',
         help='sets the channels currency name')
async def cmd_set_currency_name(msg: Message, *args):
    if len(args) != 1:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_set_currency_name)

    set_currency_name(msg.channel_name, args[0])
    await msg.reply(
        f"this channel's currency name is now \"{get_currency_name(msg.channel_name).name}\"")


@Command('getcurrencyname', help='get the channels current currency name')
async def cmd_get_currency_name(msg: Message, *ignored):
    await msg.reply(
        f'this channel\'s current currency name is "{get_currency_name(msg.channel_name).name}"')


@Command('bal', syntax='(target)', help='gets the caller\'s (or target\'s if specified) balance')
async def cmd_get_bal(msg: Message, *args):
    if args:
        target = args[0].lstrip('@')

        if target not in msg.channel.chatters:
            raise InvalidArgumentsError(reason=f'no viewer found by the name of "{target}"', cmd=cmd_get_bal)
    else:
        target = msg.author

    currency_name = get_currency_name(msg.channel_name).name
    balance = get_balance(msg.channel_name, target).balance
    await msg.reply(whisper=True, msg=f'@{target} has {balance} {currency_name}')


@Command('setbal', permission=MANAGE_CURRENCY_PERMISSION, syntax='<new_balance> (target)',
         help='sets the callers or targets balance')
async def cmd_set_bal(msg: Message, *args):
    if not len(args):
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_set_bal)
    elif len(args) == 2:
        target = args[1].lstrip('@')
        if target not in msg.channel.chatters:
            raise InvalidArgumentsError(reason=f'no viewer found by the name of "{args[1]}"', cmd=cmd_set_bal)
    else:
        target = msg.author

    try:
        new_balance = int(args[0])
        if new_balance < 0:
            raise InvalidArgumentsError(reason='new balance cannot be negative', cmd=cmd_set_bal)

        set_balance(msg.channel_name, target, new_balance)
    except ValueError:
        raise InvalidArgumentsError(reason=f'target balance must be a integer. example: 100')

    await msg.reply(
        f'@{target} now has {args[0]} '
        f'{get_currency_name(msg.channel_name).name}')


@Command('addbal', permission=MANAGE_CURRENCY_PERMISSION, syntax='<user or all> <amount>',
         help='adds the balance of the target, or all')
async def cmd_add_bal(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError('must supply both <user or all> and <amount>', cmd=cmd_add_bal)

    target = args[0].lower()
    if target != 'all' and target not in msg.channel.chatters:
        raise InvalidArgumentsError(f'cannot find any viewer named "{target}"', cmd=cmd_add_bal)

    try:
        amount = int(args[1])
    except ValueError:
        raise InvalidArgumentsError(f'cannot parse "{args[1]}" to a int, must be a valid int, ex: 100', cmd=cmd_add_bal)

    if amount <= 0:
        raise InvalidArgumentsError(f'amount must be positive and not zero, ex: 1, 2, 100', cmd=cmd_add_bal)

    currency = get_currency_name(msg.channel_name).name
    if target == 'all':
        add_balance_to_all(msg.channel_name, amount)
        await msg.reply(f"added {amount} {currency} to everyone's balance")
    else:
        add_balance(msg.channel_name, target, amount)
        await msg.reply(
            f'gave {target} {amount} {currency}, '
            f'their total is now {get_balance(msg.channel_name, target).balance} {currency}')


@Command('subbal', permission=MANAGE_CURRENCY_PERMISSION, syntax='<user or all> <amount>',
         help='subtracts the balance of the target, or all')
async def cmd_sub_bal(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError('must supply both <user or all> and <amount>', cmd=cmd_sub_bal)

    target = args[0].lower()
    if target != 'all' and target not in msg.channel.chatters:
        raise InvalidArgumentsError(f'cannot find any viewer named "{target}"', cmd=cmd_sub_bal)

    try:
        amount = int(args[1])
    except ValueError:
        raise InvalidArgumentsError(f'cannot parse "{args[1]}" to a int, must be a valid int, ex: 100', cmd=cmd_sub_bal)

    if amount <= 0:
        raise InvalidArgumentsError(f'amount must be positive and not zero, ex: 1, 2, 100', cmd=cmd_sub_bal)

    currency = get_currency_name(msg.channel_name).name

    if get_balance_from_msg(msg).balance < amount:
        raise InvalidArgumentsError(f'{target} does not have {currency} to subtract {amount} {currency} from',
                                    cmd=cmd_sub_bal)

    if target == 'all':
        subtract_balance_from_all(msg.channel_name, amount)
        await msg.reply(f"subtracted {amount} {currency} from everyone's balance")
    else:
        subtract_balance(msg.channel_name, target, amount)
        await msg.reply(
            f'subtracted {amount} {currency} from {target}, '
            f'their total is now {get_balance(msg.channel_name, target).balance} {currency}')


@Command('give', syntax='<target> <amount>',
         help='gives the target the specified amount from the callers currency balance')
async def cmd_give(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_give)

    if not msg.mentions or msg.mentions[0] not in msg.channel.chatters:
        raise InvalidArgumentsError(reason=f'no viewer found by the name "{(msg.mentions or args)[0]}"')

    caller = get_balance_from_msg(msg)
    target = get_balance(msg.channel_name, msg.mentions[0])

    try:
        give = int(args[1])
    except ValueError:
        raise InvalidArgumentsError(reason='give amount must be a integer, example: 100', cmd=cmd_give)

    if give <= 0:
        raise InvalidArgumentsError(reason='give amount must be 1 or higher', cmd=cmd_give)

    cur_name = get_currency_name(msg.channel_name).name

    if caller.balance < give:
        raise InvalidArgumentsError(reason=f"{msg.mention} you don't have enough {cur_name}", cmd=cmd_give)

    caller.balance -= give
    target.balance += give

    session.commit()

    await msg.reply(
        f"@{msg.author} you gave @{args[0]} {give} {cur_name}, @{args[0]}'s balance is now {target.balance}")


@Command('gamble', syntax='<dice_sides> <bet>',
         help='throws a X sided die and if it rolls on 1 you get twice your bet + (bet*(6/<dice_sides>)), '
              'if the dice sides are more than 6 you get more payout, '
              'but it is also a lower chance to roll a 1.')
async def cmd_gamble(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_gamble)

    try:
        sides = int(args[0])
        bet = int(args[1])
    except ValueError:
        raise InvalidArgumentsError(reason='invalid value for sides or bet', cmd=cmd_gamble)

    if bet < 10:
        raise InvalidArgumentsError(reason='bet cannot be less then 10', cmd=cmd_gamble)

    elif sides < 2:
        raise InvalidArgumentsError(reason='sides cannot be less than 2', cmd=cmd_gamble)

    bal = get_balance_from_msg(msg)
    cur_name = get_currency_name(msg.channel_name).name

    if bal.balance < bet:
        raise InvalidArgumentsError(reason=f"{msg.mention} you don't have enough {cur_name}", cmd=cmd_gamble)

    n = randbelow(sides) + 1

    if n == 1:
        if sides >= 6:
            bet *= 2
        gain = bet + int(bet * (sides / 6))
        bal.balance += gain
        await msg.reply(f'you rolled {n} and won {gain} {cur_name}')

    else:
        bal.balance -= bet
        await msg.reply(f'you rolled {n} and lost your bet of {bet} {cur_name}')

    session.commit()


last_mine_time = {}
mine_gain = 50


@Command('mine', help='mines for currency, gives you a predefined amount (default 50)')
async def cmd_mine(msg: Message, *args):
    key = (msg.author, msg.channel_name)
    diff = (datetime.now() - last_mine_time.get(key, datetime.now())).total_seconds()

    if key not in last_mine_time or diff >= 0:
        bal = get_balance_from_msg(msg)
        bal.balance += mine_gain
        session.commit()
        last_mine_time[key] = datetime.now() + timedelta(minutes=5)

        await msg.reply(
            f'@{msg.author} you went to work at the mines and came out with '
            f'{mine_gain} {get_currency_name(msg.channel_name).name} worth of gold',
            whisper=True)
    else:
        await msg.reply(f'you cannot mine again for {int(abs(diff))} seconds', whisper=True)


@Command('top', help="lists the top 10 balance holders")
async def cmd_top(msg: Message, *args):
    results = (session.query(Balance)
               .filter(Balance.channel == msg.channel_name, Balance.user != msg.channel_name,
                       Balance.user != get_nick().lower())
               .order_by(Balance.balance.desc())
               .limit(10))

    b: Balance
    message = ' | '.join(f'{i}: {b.user} => {b.balance}' for i, b in enumerate(results, 1))

    await msg.reply(message or 'no users found')


running_arenas: Dict[str, Arena] = {}


@Command('arena', syntax='<entry_fee>',
         help='starts a arena match, waits a certain amount of time for ppl to enter, '
              'if not enough ppl enter the arena is cancelled and everyone is refunded,'
              'the winner gets all of the entry_fee\'s paid')
async def cmd_arena(msg: Message, *args):
    def _can_pay_entry_fee(fee):
        return get_balance(msg.channel_name, msg.author).balance >= fee

    def _remove_running_arena_entry(arena: Arena):
        try:
            del running_arenas[arena.channel.name]
        except KeyError:
            pass

    arena = running_arenas.get(msg.channel_name)
    curname = get_currency_name(msg.channel_name).name

    # arena is already running for this channel
    if arena:
        if msg.author in arena.users:
            return await msg.reply(
                whisper=True,
                msg='you are already entered the in the arena')

        elif not _can_pay_entry_fee(arena.entry_fee):
            await msg.reply(
                whisper=True,
                msg=f'{msg.mention} you do not have enough {curname} '
                    f'to join the arena, entry_fee is {arena.entry_fee} {curname}')
            return

        arena.add_user(msg.author)
        add_balance(msg.channel_name, msg.author, -arena.entry_fee)

        await msg.reply(
            whisper=True,
            msg=f'{msg.mention} you have been added to the arena, '
                f'you were charged {arena.entry_fee} {curname} for entry')

    # start a new arena as one is not already running for this channel
    else:
        if args:
            try:
                entry_fee = int(args[0])
            except ValueError:
                raise InvalidArgumentsError(reason='invalid value for entry fee, example: 100', cmd=cmd_arena)
        else:
            entry_fee = ARENA_DEFAULT_ENTRY_FEE

        if entry_fee and entry_fee < ARENA_DEFAULT_ENTRY_FEE:
            raise InvalidArgumentsError(reason=f'entry fee cannot be less than {ARENA_DEFAULT_ENTRY_FEE}',
                                        cmd=cmd_arena)

        if not _can_pay_entry_fee(entry_fee):
            await msg.reply(
                whisper=True,
                msg=f'{msg.mention} you do not have {entry_fee} {curname}')
            return

        arena = Arena(msg.channel, entry_fee, on_arena_ended_func=_remove_running_arena_entry)
        arena.start()
        arena.add_user(msg.author)

        subtract_balance(msg.channel_name, msg.author, arena.entry_fee)

        running_arenas[msg.channel_name] = arena


@Command('duel', syntax='<target_user> (amount, default: 10)',
         help='challenges a user to a duel with the bid as the reward')
async def cmd_duel(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_duel)

    target = args[0].lstrip('@')

    if target == msg.author:
        raise InvalidArgumentsError(reason='you cannot duel yourself', cmd=cmd_duel)

    if target not in msg.channel.chatters:
        raise InvalidArgumentsError(reason=f'{msg.mention} {target} is not in this channel', cmd=cmd_duel)

    duel = get_duel(msg.channel_name, msg.author, target)

    if duel and not duel_expired(duel):
        raise InvalidArgumentsError(reason=f'{msg.mention} you already have a pending duel with {target}', cmd=cmd_duel)

    try:
        bet = int(args[1])
    except ValueError:
        raise InvalidArgumentsError(reason=f'invalid bet: {args[1]}, bet must be a number with no decimals, ex: 12',
                                    cmd=cmd_duel)
    except IndexError:
        bet = 10

    add_duel(msg.channel_name, msg.author, target, bet)

    currency_name = get_currency_name(msg.channel_name).name
    await msg.reply(
        f'{msg.mention} has challenged @{target} to a duel for {bet} {currency_name}'
        f', do "{cfg.prefix}accept {msg.mention}" to accept the duel')


@Command('accept', syntax='<challenger>', help='accepts a duel issued by the challenger that is passed to this command')
async def cmd_accept(msg: Message, *args):
    if len(args) != 1:
        raise InvalidArgumentsError('missing required arguments')

    challenger = args[0].lstrip('@')
    winner, bet = accept_duel(msg.channel_name, challenger, msg.author)

    if not winner:
        raise InvalidArgumentsError(
            reason=f'{msg.mention}, you have not been challenged by {challenger}, or the duel might have expired',
            cmd=cmd_accept)

    loser = msg.author if winner == msg.author else challenger

    add_balance(msg.channel_name, winner, bet)
    subtract_balance(msg.channel_name, loser, bet)

    currency_name = get_currency_name(msg.channel_name).name
    await msg.reply(f'@{winner} has won the duel, {bet} {currency_name} went to the winner')



from asyncio import get_event_loop
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..message import Message

__all__ = ['run_command']


async def run_command(name: str, msg: 'Message', args: List[str] = None, blocking: bool = True):
    """
    runs a command from another command, can be used for chaining command pragmatically,
    the base message passed to this function MUST be a WHISPER or PRIVMSG for this to work

    example (called from another command)

    >>> # `msg` is from the first command triggered from the actual chat from the channel
    >>> # ['17'] is the arguments passed to this newly command, can be empty
    >>> await run_command('roll', msg, ['17'])

    :param name: the name of the command to call, if the name is not found, it tries it with the prefix added
    :param msg: the message instance of first command called, needed to retain state, and keep natural flow
    :param args: the arguments to pass to the newly called command
    :param blocking: if True, the caller will wait for newly called command to be done before continuing,
                     else, run it as another task and continue
    """
    from twitchbot import Message, get_command, get_custom_command, Command, CustomCommand, CustomCommandAction

    cmd: Command = get_command(name) or get_custom_command(msg.channel_name, name)
    if not cmd:
        raise ValueError(f'[run_command] could not find command or custom command by the name of "{name}"')
    if isinstance(cmd, CustomCommand):
        cmd = CustomCommandAction(cmd)

    args = [f'"{arg}"' if ' ' in arg else arg for arg in (args or [])]
    raw = msg.raw_msg
    # get the original info line from the twitch message, then append our new arguments to it to be parsed
    text = raw[:raw.index(':', raw.index(':') + 1) + 1]
    # create a new message from the formatted new raw text, this is needed to ensure everything flows as intended
    # here the base message data is used, then we replace the context with our own command and arguments
    new_msg = Message(f"{text}{cmd.fullname} {' '.join(args)}", irc=msg.irc, bot=msg.bot)

    if blocking:
        await cmd.execute(new_msg)
    else:
        get_event_loop().create_task(cmd.execute(new_msg))



from typing import NamedTuple

__all__ = [
    'RateLimit'
]


class RateLimit(NamedTuple):
    remaining: int
    limit: int
    reset: int

    @staticmethod
    def from_headers(headers):
        return RateLimit(
            limit=int(headers.get('Ratelimit-Limit')),
            remaining=int(headers.get('Ratelimit-Remaining')),
            reset=int(headers.get('Ratelimit-Reset'))
        )



from typing import Union, Optional
from .models import Quote
from .session import session

__all__ = ('quote_exist', 'add_quote', 'get_quote', 'get_quote_by_alias', 'get_quote_by_id', 'delete_all_quotes',
           'delete_quote_by_alias', 'delete_quote_by_id')


def quote_exist(channel: str, id: int = None, alias: str = None) -> bool:
    """return if quote exist that has the same ID or ALIAS or both"""

    if id is None and alias is None:
        return False

    filters = [Quote.channel == channel]

    if id is not None:
        filters.append(Quote.id == id)
    if alias is not None:
        filters.append(Quote.alias == alias)

    return bool(session.query(Quote).filter(*filters).count())


def add_quote(quote: Quote) -> bool:
    """adds a quote to the quote DB, return a bool indicating if it was successful"""
    assert isinstance(quote, Quote), 'quote must of type Quote'

    if quote_exist(quote.channel, quote.id, quote.alias):
        return False

    session.add(quote)
    session.commit()
    return True


def get_quote_by_id(channel: str, id: int) -> Optional[Quote]:
    assert isinstance(id, int), 'quote_id must be of type int'
    return session.query(Quote).filter(Quote.id == id, Quote.channel == channel).one_or_none()


def get_quote_by_alias(channel: str, alias: str) -> Optional[Quote]:
    assert isinstance(alias, str), 'quote_alias must be of type str'
    return session.query(Quote).filter(Quote.alias == alias, Quote.channel == channel).one_or_none()


def get_quote(channel: str, id_or_alias: Union[str, int]) -> Optional[Quote]:
    """
    tries to find quote by parsing x to int first (uses value if its already a int),
    then tries to find quote using x as a alias
    returns the quote if one exist, else None
    """
    try:
        return get_quote_by_id(channel, int(id_or_alias))
    except ValueError:
        return get_quote_by_alias(channel, str(id_or_alias))


def delete_quote_by_id(channel: str, id: int) -> None:
    assert isinstance(id, int), 'quote_id must be of type int'
    session.query(Quote).filter(Quote.channel == channel, Quote.id == id).delete()
    session.commit()


def delete_quote_by_alias(channel: str, alias: str) -> None:
    assert isinstance(alias, str), 'quote_alias must be of type str'
    session.query(Quote).filter(Quote.channel == channel, Quote.alias == alias).delete()
    session.commit()


def delete_all_quotes():
    session.query(Quote).delete()
    session.commit()



from twitchbot import (
    set_message_timer_active,
    set_message_timer,
    set_message_timer_interval,
    set_message_timer_message,
    restart_message_timer,
    message_timer_exist,
    delete_message_timer,
    get_all_channel_timers,
    get_message_timer,
    cfg,
    Message,
    Command,
    InvalidArgumentsError
)

PREFIX = cfg.prefix
MIN_MESSAGE_TIMER_INTERVAL = 10


async def _parse_interval(msg, value):
    try:
        interval = float(value)
        if interval < MIN_MESSAGE_TIMER_INTERVAL:
            raise ValueError()
        return True, value
    except ValueError:
        await msg.reply('invalid interval, must be a valid float and be above 10, ex: 10,  11.5, 30')
        return False, 0


@Command('addtimer', syntax='<name> <interval> <message>', help='adds a message timer')
async def cmd_add_timer(msg: Message, *args):
    if len(args) < 3:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_add_timer)

    valid, interval = await _parse_interval(msg, args[1])
    if not valid:
        return

    timer_msg = ' '.join(args[2:])
    timer_name = args[0].lower()

    if message_timer_exist(msg.channel_name, timer_name):
        raise InvalidArgumentsError(reason=f'a timer already exist by the name of "{timer_name}"', cmd=cmd_add_timer)

    set_message_timer(msg.channel_name, timer_name, timer_msg, interval)
    await msg.reply(f'created timer successfully')


@Command('starttimer', syntax='<name>', help='starts a message timer')
async def cmd_start_timer(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError('missing required argument', cmd=cmd_start_timer)

    name = args[0].lower()
    timer = get_message_timer(msg.channel_name, name)

    if not timer:
        raise InvalidArgumentsError(reason=f'no timer was found by "{name}"', cmd=cmd_start_timer)

    if timer.running:
        await msg.reply(f'timer "{name}" is already running')
        return

    if set_message_timer_active(msg.channel_name, name, True):
        await msg.reply(f'successfully started the timer "{name}"')
    else:
        await msg.reply(f'failed to start the timer "{name}"')


@Command('stoptimer', syntax='<name>', help='stops a message timer')
async def cmd_stop_timer(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_stop_timer)

    name = args[0].lower()
    timer = get_message_timer(msg.channel_name, name)

    if not timer:
        raise InvalidArgumentsError(reason=f'no timer was found by "{name}"', cmd=cmd_stop_timer)

    if not timer.running:
        await msg.reply(f'that timer is not running')
        return

    if set_message_timer_active(msg.channel_name, name, False):
        await msg.reply(f'successfully stopped the timer "{name}"')
    else:
        await msg.reply(f'failed to stop the timer "{name}"')


@Command('deltimer', syntax='<name>', help='deletes a message timer')
async def cmd_del_timer(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_del_timer)

    name = args[0].lower()
    timer = get_message_timer(msg.channel_name, name)

    if not timer:
        raise InvalidArgumentsError(reason=f'no timer was found by "{name}"', cmd=cmd_del_timer)

    if delete_message_timer(msg.channel_name, name):
        await msg.reply(f'successfully deleted timer "{name}"')
    else:
        await msg.reply(f'failed to delete timer "{name}"')


@Command('listtimers', help='lists all message timers for a channel')
async def cmd_list_timers(msg: Message, *args):
    timers = get_all_channel_timers(msg.channel_name)
    active_timers = ', '.join(timer.name for timer in timers if timer.active)
    inactive_timers = ', '.join(timer.name for timer in timers if not timer.active)

    if not active_timers and not inactive_timers:
        await msg.reply(f'no timers found for this channel')
        return

    if active_timers:
        await msg.reply(f'active timers: {active_timers}')

    if inactive_timers:
        await msg.reply(f'inactive timers: {inactive_timers}')


@Command('edittimer', syntax='<name> <msg or interval> <new value>', help="edits a timer's message or interval")
async def cmd_edit_timer(msg: Message, *args):
    if len(args) < 3:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_edit_timer)

    name = args[0].lower()
    timer = get_message_timer(msg.channel_name, name)

    if not timer:
        raise InvalidArgumentsError(reason=f'no timer was found by "{name}"', cmd=cmd_edit_timer)

    mode = args[1].lower()
    if mode not in ('msg', 'interval'):
        raise InvalidArgumentsError(reason='invalid option, must be `msg` or `interval`', cmd=cmd_edit_timer)

    if mode == 'interval':
        try:
            interval = int(args[2])
        except ValueError:
            raise InvalidArgumentsError(reason='interval must be a valid integer, ex: 10', cmd=cmd_edit_timer)

        set_message_timer_interval(msg.channel_name, name, interval)
        restart_message_timer(msg.channel_name, name)

        return await msg.reply(f'updated timer interval for "{name}"')

    elif mode == 'msg':
        value = ' '.join(args[2:])

        set_message_timer_message(msg.channel_name, name, value)
        restart_message_timer(msg.channel_name, name)

        return await msg.reply(f'updated timer message for "{name}"')



from secrets import choice
from random import randint

from twitchbot import (
    Message,
    CommandContext,
    Command,
    InvalidArgumentsError
)


@Command('roll', context=CommandContext.BOTH, syntax='(sides)', help='rolls a X sided die')
async def cmd_roll(msg: Message, *args):
    try:
        sides = int(args[0]) if args else 6
    except ValueError:
        raise InvalidArgumentsError(reason='invalid value for sides', cmd=cmd_roll)

    num = randint(1, sides)
    user = msg.mention if msg.is_privmsg else ''
    await msg.reply(f'{user} you rolled a {num}')


@Command('crashcode', permission='crashcode')
async def cmd_crash_code(msg: Message, *args):
    await msg.reply(f'you may not crash me! {msg.mention}')


@Command('choose', syntax='<option> <option> ect', help='chooses a random option passed to the command')
async def cmd_choose(msg: Message, *args):
    if len(args) < 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_choose)

    await msg.reply(f'result: {choice(args)}')


@Command('color', permission='color', syntax='<color>', help='sets the bots chat color')
async def cmd_color(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_color)

    await msg.channel.color(args[0])
    await msg.reply(f'set color to {args[0]}')


magic_8_ball_messages = [
    'it is certain',
    'yes',
    'not happening',
    'no',
    'maybe',
    "it's possible"
]


@Command('8ball', syntax='<question>', help='prints a random answer to the question')
async def cmd_8ball(msg: Message, *args):
    await msg.reply(f'{msg.mention} {choice(magic_8_ball_messages)}')



import warnings


class Tags:
    def __init__(self, tags: str):
        self.all_tags = {name.strip().replace(' ', ''): value for name, value in _split_tags(tags)}
        self.badges: dict = _parse_badges(self.all_tags.get('badges'))
        self.color: str = self.all_tags.get('color')
        self.display_name: str = self.all_tags.get('display-name')
        self.emotes: str = self.all_tags.get('emotes')
        self.id: str = self.all_tags.get('id')
        self.mod: int = _try_parse_int(self.all_tags.get('mod'))
        self.room_id: int = _try_parse_int(self.all_tags.get('room-id'))
        self.subscriber: int = _try_parse_int(self.all_tags.get('subscriber'))
        self.tmi_sent_ts: int = _try_parse_int(self.all_tags.get('tmi-sent-ts'))
        self.user_id: int = _try_parse_int(self.all_tags.get('user-id'))
        self.user_type: int = self.all_tags.get('user-type')
        self.bits: int = _try_parse_int(self.all_tags.get('bits'))
        self.bits_leader: int = self.all_tags.get('bits-leader')
        self.broadcaster: int = self.badges.get('broadcaster', 0)
        self.msg_id: str = self.all_tags.get('msg-id')
        self.raid_viewer_count: int = _try_parse_int(self.all_tags.get('msg-param-viewerCount'))
        self.resub_months: int = 0
        self.sub_plan: int = 0
        self.sub_recipient: str = self.all_tags.get('msg-param-recipient-display-name')

        # twitch sends months in different tags based on event, find the actual amount of months here
        if self.all_tags.get('msg-param-cumulative-months') is not None:
            self.resub_months = _try_parse_int(self.all_tags.get('msg-param-cumulative-months'))
        else:
            self.resub_months = _try_parse_int(self.all_tags.get('msg-param-months'))

        # attempt to figure out the person's subplan
        if self.all_tags.get('msg-param-sub-plan') != 'Prime':
            self.sub_plan = _try_parse_int(self.all_tags.get('msg-param-sub-plan'))
        else:
            # arbitrary number to signify prime status
            self.sub_plan = 500

        # bit_leader is initially a string, it is then parsed into a int here
        if self.bits_leader:
            self.bits_leader = _try_parse_int(self.bits_leader.partition('/')[-1])

    @property
    def turbo(self):
        warnings.warn('turbo is moving to badges in later twitch api versions')
        return _try_parse_int(self.all_tags.get('turbo'))

    @property
    def is_gift_sub(self):
        return self.msg_id in {'subgift', 'anonsubgift', 'submysterygift'}

    @property
    def is_sub_upgrade(self):
        return self.msg_id in {'anongiftpaidupgrade', 'giftpaidupgrade'}


def _split_tags(tags: str):
    for tag in tags.split(';'):
        name, _, value = tag.partition('=')
        yield name.replace('@', ''), value.strip()


def _parse_badges(badges: str):
    if not badges:
        return {}

    all_badges = badges.split(',')
    ret = {}

    for badge in all_badges:
        if '/' in badge:
            name, value = badge.split('/')

            if value.isdigit():
                value = int(value)

            ret[name] = value

        else:
            ret[badge] = badge

    return ret


def _try_parse_int(value, default=0):
    try:
        return int(value)
    except (ValueError, TypeError):
        return default

# example tags (split by =)
# (('@badges', 'broadcaster/1'),
# ('color', '#FF69B4'),
# ('display-name', 'bob'),
# ('emotes', ''),
# ('id', '4585b203-ad2e-40ab-9a54-e4d6e91cb85e'),
# ('mod', '0'),
# ('room-id', '1234'),
# ('subscriber', '0'),
# ('tmi-sent-ts', '1527291908857'),
# ('turbo', '0'),
# ('user-id', '1234'),
# ('user-type', ' '))



from typing import Optional, TYPE_CHECKING

__all__ = [
    'get_bot', 'set_bot'
]

if TYPE_CHECKING:
    from .bots import BaseBot

_bot: Optional['BaseBot'] = None


def get_bot() -> Optional['BaseBot']:
    return _bot


def set_bot(bot: 'BaseBot'):
    global _bot
    _bot = bot



from typing import Optional, List

from .session import session
from .models import CustomCommand

__all__ = (
    'custom_command_exist',
    'get_custom_command',
    'add_custom_command',
    'delete_custom_command',
    'get_all_custom_commands',
)


def custom_command_exist(channel: str, name: str) -> bool:
    return bool(
        session.query(CustomCommand).filter(CustomCommand.channel == channel, CustomCommand.name == name).count())


def get_custom_command(channel: str, name: str) -> Optional[CustomCommand]:
    """gets a custom command from the DB, returns the command if found, else None"""
    assert isinstance(name, str), 'name must be of type str'
    return session.query(CustomCommand).filter(CustomCommand.channel == channel,
                                               CustomCommand.name == name).one_or_none()


def add_custom_command(cmd: CustomCommand) -> bool:
    """adds a custom command, returns a bool if it was successful"""

    if get_custom_command(cmd.channel, cmd.name) is not None:
        return False

    session.add(cmd)
    session.commit()
    return True


def delete_custom_command(channel: str, name: str) -> bool:
    """deletes the custom command from the DB if it exist, return if it was successful"""
    assert isinstance(name, str), 'name must be of type str'

    if get_custom_command(channel, name) is None:
        return False

    session.query(CustomCommand).filter(CustomCommand.channel == channel, CustomCommand.name == name).delete()
    session.commit()
    return True


def get_all_custom_commands(channel: str) -> List[CustomCommand]:
    return session.query(CustomCommand).filter(CustomCommand.channel == channel).all()



__all__ = ['auto_register']


def auto_register(obj):
    """
    registers the mod when the files is loaded

    example:
    >>> from twitchbot import Message
    >>>
    >>> @auto_register
    >>> class MyMod(Mod):
    >>>     async def on_privmsg_received(self, msg: Message):
    >>>         print(f'MyMod got the message: {msg.content}')

    :param obj:
    :return:
    """
    # getting around circular import issues
    from ..modloader import Mod, register_mod

    if issubclass(obj, Mod):
        register_mod(obj())
    elif isinstance(obj, Mod):
        register_mod(obj)
    else:
        raise ValueError(f'Unsupported auto_register class: {obj} type: {type(obj)}')

    return obj



import asyncio
from asyncio import Future, ensure_future
from twitchbot.channel import Channel
from secrets import choice
from twitchbot.database import add_balance, get_currency_name
from .config import cfg

ARENA_WAIT_TIME = 30
ARENA_DEFAULT_ENTRY_FEE = 30

VICTORY_MESSAGES = [
    '{winner} fought long and hard, and won {winnings} {currency}',
    '{winner} has won the FFA, and walked away with {winnings} {currency}!',
]


class Arena:
    def __init__(self, channel, entry_fee=ARENA_DEFAULT_ENTRY_FEE, min_users=2, on_arena_ended_func=None):
        self.on_arena_ended_func = on_arena_ended_func
        self.future: Future = None
        self.channel: Channel = channel

        self.entry_fee = entry_fee
        self.min_users = min_users
        self.users = set()
        self.running = False

    async def _start_countdown(self, delay):
        curname = get_currency_name(self.channel.name).name

        await self.channel.send_message(
            f'FFA arena has been opened and will start in {delay} seconds! Entry fee is {self.entry_fee} {curname}. '
            f'Type {cfg.prefix}arena to join')

        await asyncio.sleep(delay)
        await self._start_arena()

    async def _start_arena(self):
        if len(self.users) < self.min_users:
            await self.channel.send_message(
                f'not enough users joined the arena to start, everyone that entered was issued a refund')

            for user in self.users:
                add_balance(self.channel.name, user, self.entry_fee)

        else:
            currency = get_currency_name(self.channel.name).name
            winner = choice(tuple(self.users))
            winnings = self.entry_fee * len(self.users)

            add_balance(self.channel.name, winner, winnings)
            await self.channel.send_message(
                choice(VICTORY_MESSAGES).format(winner=winner, winnings=winnings, currency=currency))

            self.users.clear()

        if self.on_arena_ended_func:
            self.on_arena_ended_func(self)

        self.running = False

    def add_user(self, user: str):
        if not self.running:
            return False

        self.users.add(user)
        return True

    def start(self):
        self.running = True
        self.future = ensure_future(self._start_countdown(ARENA_WAIT_TIME))



import asyncio
import ssl

from ..config import get_nick, get_oauth
from ..irc import Irc

__all__ = ('SSL_PORT', 'HTTP_PORT', 'create_connection', 'send_auth', 'create_irc')

SSL_PORT = 443
HTTP_PORT = 6667


async def create_connection():
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    ssl_context.get_ciphers()
    return await asyncio.open_connection('irc.chat.twitch.tv', SSL_PORT, ssl=ssl_context)


def send_auth(irc: Irc):
    irc.send_all(
        f'PASS {get_oauth()}',
        f'NICK {get_nick()}')


async def create_irc() -> Irc:
    reader, writer = await create_connection()
    return Irc(reader=reader, writer=writer)



from asyncio import sleep, ensure_future
from typing import Optional, Dict, List

from .models import MessageTimer
from .session import session
from ..channel import channels

__all__ = ('get_message_timer', 'set_message_timer', 'message_timer_exist', 'set_message_timer_interval',
           'set_message_timer_message', 'delete_all_message_timers', 'delete_message_timer', 'set_message_timer_active',
           'active_message_timers', 'get_all_channel_timers', 'restart_message_timer')

active_message_timers: Dict[str, MessageTimer] = {}


def timer_one_or_none(channel, *criteria) -> Optional[MessageTimer]:
    return session.query(MessageTimer).filter(MessageTimer.channel == channel, *criteria).one_or_none()


def get_message_timer(channel: str, name: str) -> Optional[MessageTimer]:
    """gets a MessageTimer instance from the database, return the MessageTimer if one is found, else None"""
    return timer_one_or_none(channel, MessageTimer.name == name)


def get_all_channel_timers(channel: str) -> List[MessageTimer]:
    return session.query(MessageTimer).filter(MessageTimer.channel == channel).all()


def set_message_timer(channel: str, name: str, message: str, interval: float, active=False) -> None:
    """updates or adds a MessageTimer to the database"""
    timer = get_message_timer(channel, name)

    if timer:
        timer.message = message
        timer.interval = interval
    else:
        timer = MessageTimer.create(channel, name, message, interval)
        session.add(timer)

    session.commit()


def set_message_timer_interval(channel: str, name: str, interval: float) -> bool:
    """updates a MessageTimers interval, returns a bool if it was successful"""
    timer = get_message_timer(channel, name)

    if not timer:
        return False

    timer.interval = interval
    session.commit()
    return True


def set_message_timer_message(channel: str, name: str, message: str) -> bool:
    """updates a MessageTimers message, returns a bool if it was successful"""
    timer = get_message_timer(channel, name)

    if not timer:
        return False

    timer.message = message
    session.commit()
    return True


def set_message_timer_active(channel: str, name: str, value: bool) -> bool:
    """sets the timers active state in the database,
    if its already running it updates it as well,
    returns if it was successful"""

    key = _key(channel, name)
    timer = get_message_timer(channel, name)

    if not timer:
        return False

    if value:
        _start_message_timer(channel, name)
    else:
        _stop_message_timer(channel, name)

    timer.active = value
    session.commit()

    return True


def message_timer_exist(channel: str, name: str) -> bool:
    """checks if a timer exists, returns a bool"""
    return bool(session.query(MessageTimer).filter(MessageTimer.channel == channel, MessageTimer.name == name).count())


def delete_all_message_timers(channel: str):
    session.query(MessageTimer).filter(MessageTimer.channel == channel).delete()
    session.commit()


def delete_message_timer(channel: str, name: str) -> bool:
    """deletes a timers on the datebase, returns if it was successful"""
    timer = get_message_timer(channel, name)

    if not timer:
        return False

    if _key(channel, name) in active_message_timers:
        _stop_message_timer(channel, name)

    session.delete(timer)
    session.commit()
    return True


# fixme: bug with restarting a active message timer?
def restart_message_timer(channel: str, name: str):
    if _key(channel, name) in active_message_timers:
        set_message_timer_active(channel, name, False)

    set_message_timer_active(channel, name, True)


def _start_message_timer(channel: str, name: str) -> bool:
    """starts a message timer, returns if it was successful"""
    key = _key(channel, name)
    timer = active_message_timers.get(key)

    if not timer:
        timer = get_message_timer(channel, name)

    if timer is None or timer.running:
        return False

    channel = channels.get(channel)

    if not channel:
        return False

    if not timer.running:
        timer.task = ensure_future(_message_timer_say_loop(channel, timer))

    active_message_timers[key] = timer
    return True


def _stop_message_timer(channel: str, name: str) -> bool:
    """stops a message timer, returns if it was successful"""
    key = _key(channel, name)
    timer = active_message_timers.get(key)

    if not timer:
        return False

    if not timer.running or timer.task is None:
        return False

    timer.task.cancel()
    del active_message_timers[key]

    return True


async def _message_timer_say_loop(channel, timer):
    while True:
        await sleep(timer.interval)
        await channel.send_message(timer.message)


def _key(channel, name):
    return f'{channel}_{name}'



from pathlib import Path
from .config import Config
from asyncio import get_event_loop

__all__ = ('cfg_disabled_mods', 'disable_mod', 'enable_mod', 'is_mod_disabled')


def _ensure_channel_data_exists(channel: str):
    """
    creates a channel's config data if it does not exist
    :param channel: the channel to verify data exist for
    """
    if channel not in cfg_disabled_mods.data:
        cfg_disabled_mods[channel] = []


def is_mod_disabled(channel: str, mod: str) -> bool:
    """
    checks if a mod is disabled for a channel
    :param channel: the channel to check for if the mod is disabled in
    :param mod: the mod to check if it is disabled for the channel
    :return: bool indicating if the mod is disabled for the channel
    """
    if channel not in cfg_disabled_mods.data:
        return False

    return mod in cfg_disabled_mods[channel]


def disable_mod(channel: str, mod: str):
    """
    disables a mod for the given channel
    :param channel: the channel to disable the mod for
    :param mod: the mod to disable for the channel
    """
    # "hack" to avoid circular import
    from .modloader import mods

    _ensure_channel_data_exists(channel)

    if mod not in cfg_disabled_mods[channel]:
        cfg_disabled_mods[channel].append(mod)
        cfg_disabled_mods.save()

    get_event_loop().create_task(mods[mod].on_disable(channel))


def enable_mod(channel: str, mod: str):
    """
    enables a mod for the given channel
    :param channel: channel to enable the mod for
    :param mod: the mod to enable for the channel
    """
    # "hack" to avoid circular import
    from .modloader import mods

    _ensure_channel_data_exists(channel)

    if mod not in cfg_disabled_mods[channel]:
        return

    cfg_disabled_mods[channel].remove(mod)
    cfg_disabled_mods.save()
    get_event_loop().create_task(mods[mod].on_enable(channel))


cfg_disabled_mods = Config(Path('configs', 'disabled_mods.json'))



import os
import typing
from datetime import datetime
from importlib import import_module
from typing import Dict, Callable, Optional, List, Tuple

from twitchbot.database import CustomCommand
from twitchbot.message import Message
from .config import cfg
from .enums import CommandContext
from .util import get_py_files, get_file_name
from .util import temp_syspath

if typing.TYPE_CHECKING:
    from .modloader import Mod

DEFAULT_COOLDOWN_BYPASS = 'bypass_cooldown'
DEFAULT_COOLDOWN = 0

__all__ = (
    'Command', 'commands', 'command_exist', 'load_commands_from_directory', 'DummyCommand', 'CustomCommandAction',
    'ModCommand', 'SubCommand', 'get_command', 'CUSTOM_COMMAND_PLACEHOLDERS', 'command_last_execute',
    'get_time_since_execute', 'reset_command_last_execute', 'is_command_off_cooldown', 'is_command_on_cooldown',
    'update_command_last_execute')


class Command:
    def __init__(self, name: str, prefix: str = None, func: Callable = None, global_command: bool = True,
                 context: CommandContext = CommandContext.CHANNEL, permission: str = None, syntax: str = None,
                 help: str = None, aliases: List[str] = None, cooldown: int = DEFAULT_COOLDOWN,
                 cooldown_bypass: str = DEFAULT_COOLDOWN_BYPASS):
        """
        :param cooldown: time between when this command when can be run, 0 means the command be run without any delay and is default value
        :param syntax: help message for how to use the command, <> is required, () is optional
        :param permission: permission needed to run the command in chat
        :param help: help message for the command, used with the `help` command
        :param name: name of the command (without the prefix)
        :param prefix: prefix require before the command name (defaults the the configs prefix if None)
        :param func: the function that the commands executes
        :param global_command: should the command be registered globally?
        :param context: the context through which calling the command is allowed
        :param aliases: aliases for this same command, only works if global_command is True
        """
        self.cooldown_bypass = cooldown_bypass
        self.cooldown: int = cooldown
        self.aliases: List[str] = aliases if aliases is not None else []
        self.help: str = help
        self.syntax: str = syntax
        self.permission: str = permission
        self.context: CommandContext = context
        self.prefix: str = (prefix if prefix is not None else cfg.prefix).lower()
        self.func: Callable = func
        self.name: str = name.lower()
        self.fullname: str = self.prefix + self.name
        self.sub_cmds: Dict[str, Command] = {}
        self.parent: Command = None

        if global_command:
            commands[self.fullname] = self

            # register all aliases passed to this functions
            if aliases is not None:
                for alias in aliases:
                    commands[self.prefix + alias] = self

    def _get_cmd_func(self, args) -> Tuple['Callable', List[str]]:
        """returns a tuple of the final commands command function and the remaining argument"""
        if not self.sub_cmds or not args or args[0].lower() not in self.sub_cmds:
            return self.func, args

        return self.sub_cmds[args[0].lower()]._get_cmd_func(args[1:])

        # while verison:
        # cmd = self
        # while cmd.sub_cmds and args and args[0].lower() in cmd.sub_cmds:
        #     cmd = cmd.sub_cmds[args[0].lower()]
        #     args = args[1:]
        #
        # return cmd.func, args

    async def execute(self, msg: Message):
        func, args = self._get_cmd_func(msg.parts[1:])
        await func(msg, *args)

    # decorator support
    def __call__(self, func) -> 'Command':
        self.func = func
        return self

    def __str__(self):
        return f'<{self.__class__.__name__} fullname={repr(self.fullname)} parent={self.parent}>'

    def __getitem__(self, item):
        return self.sub_cmds.get(item.lower()) or self.sub_cmds.get(item.lower()[1:])


class SubCommand(Command):
    def __init__(self, parent: Command, name: str, func: Callable = None, permission: str = None, syntax: str = None,
                 help: str = None, cooldown: int = DEFAULT_COOLDOWN, cooldown_bypass: str = DEFAULT_COOLDOWN_BYPASS):
        super().__init__(name=name, prefix='', func=func, permission=permission, syntax=syntax, help=help,
                         global_command=False, cooldown=cooldown, cooldown_bypass=cooldown_bypass)

        self.parent: Command = parent
        self.parent.sub_cmds[self.name] = self


class DummyCommand(Command):
    def __init__(self, name: str, prefix: str = None, global_command: bool = True,
                 context: CommandContext = CommandContext.CHANNEL, permission: str = None, syntax: str = None,
                 help: str = None):
        super().__init__(name=name, prefix=prefix, func=self.exec, global_command=global_command,
                         context=context, permission=permission, syntax=syntax, help=help)

    async def exec(self, msg: Message, *args):
        """the function called when the dummy command is executed"""
        if self.sub_cmds:
            await msg.reply(f'command options: {", ".join(self.sub_cmds)}')
        else:
            await msg.reply('no sub-commands were found for this command')

    def add_sub_cmd(self, name: str) -> 'DummyCommand':
        """adds a new DummyCommand to the current DummyCommand as a sub-command, then returns the new DummyCommand"""
        cmd = DummyCommand(name, prefix='', global_command=False)
        self.sub_cmds[cmd.fullname] = cmd
        return cmd


def _calc_channel_live_time(msg) -> str:
    if msg.channel.live:
        return format((msg.channel.stats.started_at - datetime.now()).total_seconds() / 3600, '.1f')

    return '[NOT LIVE]'


CUSTOM_COMMAND_PLACEHOLDERS = (
    (
        '%user',
        lambda msg: f'@{msg.author}'
    ),
    (
        '%uptime',
        _calc_channel_live_time
    ),
    (
        '%channel',
        lambda msg: msg.channel_name
    ),
)


class CustomCommandAction(Command):
    def __init__(self, cmd):
        super().__init__(cmd.name, prefix='', func=self.execute, global_command=False)
        self.cmd: CustomCommand = cmd
        self.cooldown = 0

    async def execute(self, msg: Message):
        resp = self.cmd.response

        for placeholder, func in CUSTOM_COMMAND_PLACEHOLDERS:
            if placeholder in resp:
                resp = resp.replace(placeholder, func(msg))

        await msg.channel.send_message(resp)


class ModCommand(Command):
    def __init__(self, mod_name: str, name: str, prefix: str = None, func: Callable = None, global_command: bool = True,
                 context: CommandContext = CommandContext.CHANNEL, permission: str = None, syntax: str = None,
                 help: str = None, cooldown: int = DEFAULT_COOLDOWN, cooldown_bypass: str = DEFAULT_COOLDOWN_BYPASS):
        super().__init__(name=name, prefix=prefix, func=func, global_command=global_command, context=context,
                         permission=permission, syntax=syntax, help=help, cooldown=cooldown,
                         cooldown_bypass=cooldown_bypass)
        self.mod_name = mod_name

    @property
    def mod(self):
        from .modloader import mods
        return mods[self.mod_name]

    async def execute(self, msg: Message):
        func, args = self._get_cmd_func(msg.parts[1:])
        if 'self' in func.__code__.co_varnames:
            await func(self.mod, msg, *args)
        else:
            await func(msg, *args)


commands: Dict[str, Command] = {}
command_last_execute: Dict[Tuple[str, str], datetime] = {}


def _create_cooldown_key(channel: str, cmd: str) -> Tuple[str, str]:
    return channel.lower(), cmd.lower()


def is_command_off_cooldown(channel: str, cmd: str, cooldown: int = None) -> bool:
    if not command_exist(cmd):
        return True
    return get_time_since_execute(channel, cmd) >= (cooldown or get_command(cmd).cooldown)


def is_command_on_cooldown(channel: str, cmd: str, cooldown: int = None) -> bool:
    return not is_command_off_cooldown(channel, cmd, cooldown)


def get_time_since_execute(channel: str, cmd: str) -> int:
    last_execute = command_last_execute.get(_create_cooldown_key(channel, cmd), datetime.min)
    return int(abs((last_execute - datetime.now()).total_seconds()))


def update_command_last_execute(channel: str, cmd: str):
    command_last_execute[_create_cooldown_key(channel, cmd)] = datetime.now()


def reset_command_last_execute(channel: str, cmd: str):
    command_last_execute[_create_cooldown_key(channel, cmd)] = datetime.min


def load_commands_from_directory(path):
    print(f'loading commands from {path}')

    path = os.path.abspath(path)

    if not os.path.exists(path):
        return

    with temp_syspath(path):
        for file in get_py_files(path):
            fname = get_file_name(file)
            mod = import_module(fname)


def command_exist(name: str) -> bool:
    """
    returns a bool indicating if a command exists,
    tries added a configs prefix to the name if not found initially,
    does not check for custom commands
    """
    return any(cmd in commands for cmd in (name, cfg.prefix + name))


def get_command(name: str) -> Optional[Command]:
    """
    gets a commands,
    tries added a configs prefix to the name if not found initally,
    returns None if not exist, does not get custom commands
    """
    return commands.get(name) or commands.get(cfg.prefix + name)



import asyncio
import typing
from datetime import datetime
from typing import Dict

from .api import StreamInfoApi
from .api.chatters import Chatters
from .config import get_nick, get_client_id
from .data import UserFollowers
from .irc import Irc
from .permission import perms
from .shared import get_bot
from .util import get_user_followers, get_headers

if typing.TYPE_CHECKING:
    from .bots import BaseBot


class Channel:
    def __init__(self, name, irc, register_globally=True):
        self.irc: Irc = irc
        self.name: str = name
        self.chatters: Chatters = Chatters(self.name)
        self.is_vip: bool = False
        self.is_mod: bool = False
        self.stats: StreamInfoApi = StreamInfoApi(get_client_id(), self.name)
        self.bot: 'BaseBot' = get_bot()

        if register_globally:
            channels[self.name.lower()] = self
            perms.load_permissions(name)

    async def followers(self) -> UserFollowers:
        return await get_user_followers(self.name, get_headers())

    @property
    def live(self):
        return self.stats.started_at != datetime.min

    async def send_message(self, msg):
        await self.irc.send_privmsg(self.name, msg)

    async def send_command(self, cmd):
        await self.irc.send_privmsg(self.name, f'/{cmd}')

    # async def ban(self, user):
    #     await self.send_command(f'ban {user}')

    async def update_loop(self):
        if get_client_id() != 'CLIENT_ID':
            while True:
                await self.chatters.update()
                await self.stats.update()
                self.is_mod = get_nick().lower() in self.chatters.mods
                self.is_vip = get_nick().lower() in self.chatters.vips
                await asyncio.sleep(60)

    def start_update_loop(self):
        asyncio.get_event_loop().create_task(self.update_loop())

    async def ban(self, user: str, reason: str = ''):
        """purges a user's messages then permabans them from the channel"""
        await self.send_command(f'ban {user} {reason}')

    async def timeout(self, user: str, duration: int = 600):
        """
        purges a user's messages then timeout out (makes them unable to chat)
        for the specified duration (default 600 seconds)
        """
        await self.send_command(f'timeout {user} {duration}')

    async def purge(self, user: str):
        """
        purges a user in the room (removes their messages via a 1 second timeout)
        :param user: user to purge
        """
        await self.timeout(user, 1)

    async def color(self, color: str):
        """sets the bots chat color"""
        await self.send_command(f'color {color}')

    def __str__(self):
        return f'<Channel name={repr(self.name)}>'

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other.lower()

        if isinstance(other, Channel):
            return self.name == other.name

        return False


channels: Dict[str, Channel] = {}


# DummyChannel is a placeholder channel for when the bots sends a whisper
class DummyChannel:
    __slots__ = 'name', 'is_mod', 'is_vip'

    def __init__(self, name):
        self.name = name
        self.is_vip = False
        self.is_mod = False



from asyncio import start_server, StreamReader, StreamWriter
from traceback import format_exc

from .channel import channels, Channel
from .config import cfg
from .util import add_task, task_running, stop_task

__all__ = 'start_command_server', 'stop_command_server'

HOST = cfg.command_server_host
PORT = cfg.command_server_port
ENABLED = cfg.command_server_enabled
COMMAND_SERVER_TASK_ID = 'COMMAND_SERVER'


def start_command_server():
    if not ENABLED:
        return

    stop_command_server()

    print(f'starting command server (view host / port in config file)')
    try:
        # noinspection PyTypeChecker
        add_task(COMMAND_SERVER_TASK_ID, start_server(handle_client, HOST, PORT))
    except Exception as e:
        print(f"\n------COMMAND SERVER------\nfailed to bind/create command server\n"
              f"this does not affect the bot, but it does mean that the command console will not work/be usable\n"
              f"if this error happens a lot, command server can be disabled in the config.json in the bot's configs folder\n"
              f'\nERROR INFO: {e}\n'
              f'EXTENDED INFO: \n{format_exc()}\n\n'
              f'------COMMAND SERVER------\n')


def stop_command_server():
    if task_running(COMMAND_SERVER_TASK_ID):
        stop_task(COMMAND_SERVER_TASK_ID)


async def handle_client(reader: StreamReader, writer: StreamWriter):
    # helper function to read the next message from the client
    async def read():
        return (await reader.readline()).decode('utf8').strip()

    try:
        writer.write(b'Connected to the command server!\n')

        # wait for the client to select a valid channel to send messages
        channel_name = ''
        while channel_name not in channels:
            connected_channels = ', '.join(channels)
            writer.write(f'what channel do you want to join?\noptions: {connected_channels}\n'.encode())
            channel_name = await read()

        # client gave us a valid channel, now get the channel object from the cache
        channel: Channel = channels[channel_name]
        writer.write(b'Send `quit` to disconnect')

        # now we just relay messages send
        while True:
            msg = await read()
            if not msg or msg.lower() == 'quit':
                return

            await channel.send_message(msg)
    except ConnectionResetError:
        return



import traceback
from datetime import datetime
from itertools import islice
from typing import List, Tuple, TYPE_CHECKING, Optional, Callable, Awaitable, FrozenSet

from twitchbot import get_bot
from .util import get_message_mentions
from .channel import Channel, channels
from .irc import Irc
from .regex import RE_PRIVMSG, RE_WHISPER, RE_USER_JOIN, RE_USERNOTICE, RE_USER_PART, RE_NOTICE, RE_TIMEOUT_DURATION
from .enums import MessageType
from .util import split_message
from .tags import Tags
from .emote import emotes, Emote
from .config import cfg

if TYPE_CHECKING:
    from .bots import BaseBot


class Message:
    def __init__(self, msg, irc=None, bot=None):
        self.channel: Optional[Channel] = None
        self.author: Optional[str] = None
        self.content: Optional[str] = None
        self.parts: List[str] = []
        self.type: MessageType = MessageType.NONE
        self.raw_msg: str = msg
        self.receiver: Optional[str] = None
        self.irc: Irc = irc
        self.tags: Optional[Tags] = None
        self.emotes: List[Emote] = []
        self.mentions: Tuple[str] = ()
        self.system_message: Optional[str] = None
        self.bot: 'BaseBot' = bot
        self.reward: Optional[str] = None
        self.msg_id: Optional[str] = None
        self.timeout_seconds: Optional[int] = None

        self._parse()

    def _normalize(self, s: str):
        return s.strip().casefold()

    @property
    def normalized_parts(self) -> FrozenSet[str]:
        """
        parts of the message,
        but they are are stripped of any leading or trailing whitespace
        and converted to lowercase
        """
        return frozenset(map(self._normalize, self.parts))

    @property
    def normalized_args(self) -> FrozenSet[str]:
        """
        parts of the message starting at index 1,
        but they are are stripped of any leading or trailing whitespace
        and converted to lowercase
        """
        return frozenset(map(self._normalize, islice(self.parts, 1, None)))

    @property
    def normalized_content(self):
        """
        lowercase()'d and strip()'d version of msg.content
        """
        return self._normalize(self.content)

    @property
    def args(self) -> List[str]:
        """
        parts of the message starting at index 1
        """
        return self.parts[1:]

    def _parse(self):
        # this weird looking bit is to make sure we do not do unnecessary checks when we have found a match
        # it takes advantage of the fact that python's `or` works as true check/default value provider
        # if the first check is false, it tries the next, then if thats false, it tries the next one, and it keeps doing this
        # until one matches or it run out of functions to check
        (self._parse_usernotice()
         or self._parse_notice()
         or self._parse_privmsg()
         or self._parse_whisper()
         or self._parse_user_join()
         or self._parse_user_part()
         or self._check_ping())

        if self.parts and any(p in emotes for p in self.parts):
            self.emotes = tuple(emotes[p] for p in self.parts if p in emotes)

        if self.tags is not None and 'system-msg' in self.tags.all_tags:
            self.system_message = self.tags.all_tags['system-msg'].replace(r'\s', ' ')

        self.msg_id = (self.tags.all_tags.get('msg-id')
                       if self.tags is not None
                       else None)

    def _parse_user_part(self) -> bool:
        m = RE_USER_PART.search(self.raw_msg)
        if m:
            self.channel = channels[m['channel']]
            self.author = m['user']
            self.type = MessageType.USER_PART

        return bool(m)

    def _parse_user_join(self) -> bool:
        m = RE_USER_JOIN.search(self.raw_msg)
        if m:
            # ensure the channel exists, if it does not, create it and put it in the cache
            channel_name = m['channel']
            if channel_name not in channels:
                Channel(channel_name, irc=self.irc, register_globally=True).start_update_loop()

            self.channel = channels[channel_name]
            self.author = m['user']
            self.type = MessageType.USER_JOIN

        return bool(m)

    def _parse_whisper(self) -> bool:
        m = RE_WHISPER.search(self.raw_msg)
        if m:
            self.author = m['user']
            self.receiver = m['receiver']
            self.content = m['content']
            self.type = MessageType.WHISPER
            self.parts = split_message(self.content)
            self.channel = Channel(self.author, self.irc, register_globally=False)

        return bool(m)

    def _parse_privmsg(self) -> bool:
        m = RE_PRIVMSG.search(self.raw_msg)
        if m:
            self.channel = channels[m['channel']]
            self.author = m['user']
            self.content = m['content']
            self.type = MessageType.PRIVMSG
            self.parts = split_message(self.content)
            self.tags = Tags(m['tags'])
            self.mentions = get_message_mentions(self)

            # checking if the message contains any bit donations
            if self.tags and self.tags.bits:
                self.type = MessageType.BITS
                # bits and rewards cannot be combined, so return here
                return True

            # checking if its a channel point redemption
            self.reward = self.tags.all_tags.get('msg-id') or self.tags.all_tags.get('custom-reward-id')
            if self.reward is not None:
                self.type = MessageType.CHANNEL_POINTS_REDEMPTION

        return bool(m)

    def _parse_usernotice(self) -> bool:
        m = RE_USERNOTICE.search(self.raw_msg)
        if m:
            self.tags = Tags(m['tags'])
            self.channel = channels[m['channel']]
            self.author = self.tags.all_tags.get('login')
            self.content = m['content']
            if self.tags.msg_id in {'sub', 'resub', 'subgift', 'anonsubgift', 'submysterygift', 'anongiftpaidupgrade',
                                    'giftpaidupgrade'}:
                self.type = MessageType.SUBSCRIPTION
            elif self.tags.msg_id == 'raid':
                self.type = MessageType.RAID
                self.author = self.tags.all_tags.get('msg-param-login')
            # RAW >> @msg-id=msg_banned :tmi.twitch.tv NOTICE #X :You are permanently banned from talking in X.
            elif self.tags.msg_id == 'msg_banned':
                self.type = MessageType.BOT_PERMANENTLY_BANNED
            elif self.tags.msg_id == 'msg_timedout':
                self.type = MessageType.BOT_TIMED_OUT
            else:
                self.type = MessageType.USER_NOTICE

        return bool(m)

    def _parse_notice(self) -> bool:
        m = RE_NOTICE.search(self.raw_msg)
        if m:
            self.tags = Tags(m['tags'])
            self.channel = channels[m['channel']]
            self.content = m['content']
            if self.tags.msg_id == 'msg_banned':
                self.type = MessageType.BOT_PERMANENTLY_BANNED
            elif self.tags.msg_id == 'msg_timedout':
                self.type = MessageType.BOT_TIMED_OUT
                self.timeout_seconds = int(RE_TIMEOUT_DURATION.search(self.content)['seconds'])
            else:
                self.type = MessageType.NOTICE

        return bool(m)

    def _check_ping(self) -> bool:
        if self.raw_msg == 'PING :tmi.twitch.tv':
            self.type = MessageType.PING

        return self.type is MessageType.PING

    @property
    def is_user_message(self):
        return self.type in (MessageType.WHISPER, MessageType.PRIVMSG)

    @property
    def is_privmsg(self):
        return self.type is MessageType.PRIVMSG

    @property
    def is_whisper(self):
        return self.type is MessageType.WHISPER

    @property
    def is_subscription(self):
        return self.type is MessageType.SUBSCRIPTION

    @property
    def is_raid(self):
        return self.type is MessageType.RAID

    @property
    def mention(self):
        return f'@{self.author}' if self.author else ''

    @property
    def channel_name(self):
        if self.channel:
            return self.channel.name
        if self.author:
            return self.author
        return ''

    async def reply(self, msg: str = '', whisper=False):
        if not msg:
            raise ValueError('msg is empty, msg must be a non-empty string')

        if not isinstance(msg, str):
            msg = str(msg)

        if self.type is MessageType.PRIVMSG and (not whisper or cfg.disable_whispers):
            await self.channel.send_message(msg)

        elif self.type is MessageType.WHISPER or (whisper and self.type is MessageType.PRIVMSG):
            if self.irc is None:
                raise RuntimeError('no irc instance set for this message')

            await self.irc.send_whisper(self.author, msg)

        # used for weirder cases like NOTICE, and point redeemation, ect
        # exclude PRIVMSG and WHISPER since they are handled above as well
        elif self.type not in (MessageType.PRIVMSG, MessageType.WHISPER):
            # check we have a valid channel to send to
            if self.channel is not None and self.channel.irc is not None and self.channel.name.strip():
                # relay the message
                await self.channel.send_message(msg)

    # else:
    #     raise ValueError(f'invalid message type to reply, expected PRIVMSG or WHISPER, current: {self.type}')

    async def wait_for_reply(self, predicate: Callable[['Message'], Awaitable[bool]] = None, timeout=30, default=None,
                             raise_on_timeout=False) -> 'Message':
        """
        waits for a message matching `predicate` to be received, when its received, it returns that message.

        if no message matching `predicate` is received by the timeout, the default will be returned.

        if raise_on_timeout is True and no matching message is received, this function will raise asyncio.TimeoutError

        if raise_on_timeout is False and no matching message is received, default will be returned when it times-out

        default is by default is None
        raise_on_timeout by default is False
        predicate defaults to `same_author_predicate`
        """

        from .replywaiter import wait_for_reply, same_author_predicate

        return await wait_for_reply(predicate or same_author_predicate(self), timeout=timeout, default=default,
                                    raise_on_timeout=raise_on_timeout)

    def __str__(self):
        if self.type is MessageType.PRIVMSG:
            return f'{self.author}({self.channel.name}): {self.content}'

        elif self.type is MessageType.WHISPER:
            return f'{self.author} -> {self.receiver}: {self.content}'

        elif self.type is MessageType.PING:
            return 'PING'

        elif self.type is MessageType.SUBSCRIPTION:
            return f'{self.author} subscribed to {self.channel_name}: {self.system_message}'

        elif self.type is MessageType.RAID:
            return f'{self.author} has raided {self.channel_name} with {self.tags.raid_viewer_count} viewers!'

        elif self.type is MessageType.USER_NOTICE:
            return f'USERNOTICE for {self.author}'

        elif self.type is MessageType.USER_JOIN:
            return f'{self.author} joined {self.channel_name}'

        elif self.type is MessageType.USER_PART:
            return f'{self.author} left {self.channel_name}'

        elif self.type is MessageType.CHANNEL_POINTS_REDEMPTION:
            return f'{self.author} redeemed reward {self.reward} in #{self.channel_name}'

        elif self.type is MessageType.BITS:
            return f'{self.author} donated {self.tags.bits} bits to #{self.channel_name}'

        return self.raw_msg

    def safe_print(self):
        try:
            print(self)
        except Exception as e:
            err = traceback.format_exc()
            print(err)
            with open('printing_errors.log', 'a') as f:
                f.write(f'\n\n{datetime.now()}:\n{err}')

    def __getitem__(self, item):
        """
        if item is a slice, returns a space joined parts of that slice

        if item in a int, returns a single part at that index
        """
        if isinstance(item, int):
            return self.parts[item]

        return ' '.join(self.parts[item])

    def __len__(self):
        """
        :return: the len() of self.parts
        """
        return len(self.parts)



from pathlib import Path

from .config import Config
from .command import get_command, command_exist


def is_command_disabled(channel: str, cmd: str):
    cmd = get_command(cmd)
    if cmd is None:
        return False

    if channel in cfg_disabled_commands.data:
        return cmd.fullname in cfg_disabled_commands[channel]

    return False


def disable_command(channel: str, cmd: str):
    cmd = get_command(cmd)
    if cmd is None:
        return

    cmd_name = cmd.fullname
    if channel not in cfg_disabled_commands.data:
        cfg_disabled_commands[channel] = [cmd_name]
        return

    if cmd_name in cfg_disabled_commands[channel]:
        return

    cfg_disabled_commands[channel].append(cmd_name)
    cfg_disabled_commands.save()


def enable_command(channel: str, cmd: str):
    cmd = get_command(cmd)
    if cmd is None or channel not in cfg_disabled_commands.data:
        return

    if cmd.fullname in cfg_disabled_commands[channel]:
        cfg_disabled_commands[channel].remove(cmd.fullname)
        cfg_disabled_commands.save()


cfg_disabled_commands = Config(Path('configs', 'disabled_commands.json'))



from twitchbot import Mod, perms, Message, Command


class PermissionsMod(Mod):
    name = 'permissions'

    async def on_permission_check(self, msg: Message, cmd: Command) -> bool:
        if not cmd.permission:
            return True

        return perms.has_permission(msg.channel_name, msg.author, cmd.permission)



import re

from twitchbot import (
    Command,
    Message,
    delete_quote_by_id,
    add_quote,
    get_quote_by_alias,
    get_quote,
    Quote,
    cfg,
    InvalidArgumentsError
)

PREFIX = cfg.prefix


@Command('addquote', syntax='"<quote text>" user=(user) alias=(alias)', help='adds a quote to the database')
async def cmd_add_quote(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_add_quote)

    optionals = ' '.join(args[1:])

    user = alias = None
    if 'user=' in optionals:
        m = re.search(r'user=([\w\d]+)', msg.content)
        if not m:
            raise InvalidArgumentsError(
                reason='invalid user for user=, must be a combination of digits and letters, ex: john_doe17',
                cmd=cmd_add_quote)

        user = m.group(1)

    if 'alias=' in optionals:
        m = re.search(r'alias=([\d\w]+)', msg.content)
        if not m:
            raise InvalidArgumentsError(
                reason='invalid alias for alias=, must be a combination of digits and letters, ex: my_quote1',
                cmd=cmd_add_quote)

        alias = m.group(1)
        if get_quote_by_alias(msg.channel_name, alias) is not None:
            raise InvalidArgumentsError(reason='there is already a quote with that alias', cmd=cmd_add_quote)

    if add_quote(Quote.create(channel=msg.channel_name, value=args[0], user=user, alias=alias)):
        resp = 'successfully added quote'
    else:
        resp = 'failed to add quote, already exist'

    await msg.reply(resp)


@Command('quote', syntax='<ID or ALIAS>', help='gets a quote by ID or ALIAS')
async def cmd_get_quote(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_get_quote)

    quote = get_quote(msg.channel_name, args[0])
    if quote is None:
        raise InvalidArgumentsError(reason='no quote found', cmd=cmd_get_quote)

    await msg.reply(f'"{quote.value}" user: {quote.user} alias: {quote.alias}')


@Command('delquote', permission='delete_quote', syntax='<ID or ALIAS>', help='deletes the quote from the database')
async def cmd_del_quote(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_del_quote)

    quote = get_quote(msg.channel_name, args[0])
    if quote is None:
        raise InvalidArgumentsError(reason='no quote found', cmd=cmd_del_quote)

    delete_quote_by_id(msg.channel_name, quote.id)
    await msg.reply(f'successfully deleted quote, id: {quote.id}, alias: {quote.alias}')



import os
import sys
from contextlib import contextmanager
from datetime import datetime
from glob import iglob
from pathlib import Path
from typing import List, Union

__all__ = ('get_py_files', 'get_file_name', 'temp_syspath', 'format_datetime')


def get_py_files(path: str) -> List[str]:
    """gets all python (.py) files in a folder"""
    yield from iglob(os.path.join(path, '*.py'))


def format_datetime(dt: datetime) -> str:
    return dt.strftime('%d/%m/%Y %H:%M:%S')


def get_file_name(path: str):
    """gets the files name without the extension"""
    return os.path.basename(path).split('.')[0]


@contextmanager
def temp_syspath(fullpath: Union[str, Path]):
    """
    temporarily appends the fullpath to sys.path, yields, then removes it from sys.path
    if the fullpath is already in sys.path the append/remove is skipped
    """
    if isinstance(fullpath, Path):
        fullpath = str(fullpath)

    if not os.path.isabs(fullpath):
        fullpath = os.path.abspath(fullpath)

    if fullpath not in sys.path:
        sys.path.insert(0, fullpath)
        yield
        sys.path.remove(fullpath)
    else:
        yield



from itertools import chain
from pathlib import Path
from typing import Dict, Iterable, Tuple, Optional

from .config import Config, cfg

__all__ = ('perms', 'Permissions', 'generate_permission_files', 'permission_defaults')

permission_defaults = {
    'admin': {
        'name': 'admin',
        'permissions': ['*'],
        'members': [cfg.owner]
    },
}


class Permissions:
    def __init__(self):
        self.channels: Dict[str, Config] = {}

    def load_permissions(self, channel: str, force_update=False):
        """loads a config file (or creates the config if it doesnt exist) into the cache of the permission object"""
        channel = channel.lower()
        if channel in self and not force_update:
            return

        config = Config(
            file_path=Path(f'configs', f'{channel}_perms.json'),
            **permission_defaults)

        self.channels[channel] = config

        if channel not in config.data['admin']['members']:
            config.data['admin']['members'].append(channel)
            self.channels[channel].save()

    def iter_user_groups(self, channel: str, user: str):
        """yields all permission groups a user is in for a channel"""
        user = user.lower()
        yield from ((name, group) for name, group in self[channel] if user in group['members'])

    def iter_groups(self, channel: str):
        if channel in self:
            yield from self[channel]

    def iter_user_permissions(self, channel: str, user: str):
        """yields all permissions from a specified user for the channel"""
        user = user.lower()
        for group_name, group in self.iter_user_groups(channel, user):
            yield from group['permissions']

    def has_permission(self, channel: str, user: str, perm: str) -> bool:
        """checks if a user has a permission"""
        user, perm = user.lower(), perm.lower()
        search = {perm, '*'}
        return user == cfg.owner or any(p in search for p in self.iter_user_permissions(channel, user))

    def get_group(self, channel: str, group: str) -> Optional[dict]:
        """gets a permission group by the name passed in, returns None if not found"""
        group = group.lower()
        if channel not in self or group not in self.channels[channel].data:
            return None

        return self[channel].data[group]

    def iter_group_permissions(self, channel: str, group: str):
        group = group.lower()
        g = self.get_group(channel, group)
        if not g:
            return

        yield from g['permissions']

    def iter_group_members(self, channel: str, group: str):
        group = group.lower()
        g = self.get_group(channel, group)
        if not g:
            return

        yield from g['members']

    def add_permission(self, channel: str, group: str, perm: str) -> bool:
        """adds a permission to a permission group for a channel, returns if it was successful"""
        group, perm = group.lower(), perm.lower()
        g = self.get_group(channel, group)

        if not g:
            return False

        if perm not in g['permissions']:
            g['permissions'].append(perm)
            self[channel].save()

        return True

    def delete_permission(self, channel: str, group: str, perm: str) -> bool:
        """adds a permission to a permission group for a channel, returns if it was successful"""
        group, perm = group.lower(), perm.lower()
        g = self.get_group(channel, group)

        if not g:
            return False

        if perm in g['permissions']:
            g['permissions'].remove(perm)
            self[channel].save()

        return True

    def add_group(self, channel: str, group: str) -> bool:
        """adds a permission group to the channels config, returns if it was successful"""
        group = group.lower()
        if self.get_group(channel, group):
            return False

        self[channel].data[group] = {
            'name': group,
            'permissions': [],
            'members': []
        }
        self[channel].save()

        return True

    def delete_group(self, channel: str, group: str) -> bool:
        """adds a permission group to the channels config, returns if it was successful"""
        group = group.lower()
        if not self.get_group(channel, group):
            return False

        del self[channel].data[group]
        self[channel].save()

        return True

    def reload_permissions(self, channel=None):
        if not channel:
            for channel in tuple(self.channels):
                self.load_permissions(channel, force_update=True)
            return True

        elif channel in self:
            self.load_permissions(channel, force_update=True)
            return True

        return False

    def add_member(self, channel: str, group: str, member: str):
        group = group.lower()
        g = self.get_group(channel, group)

        if not g:
            return False

        member = member.lower()

        if member not in g['members']:
            g['members'].append(member)
            self[channel].save()

        return True

    def delete_member(self, channel: str, group: str, member: str):
        group = group.lower()
        g = self.get_group(channel, group)

        if not g:
            return False

        member = member.lower()

        if member not in g['members']:
            return False

        g['members'].remove(member)
        self[channel].save()

        return True

    def __contains__(self, item):
        return item in self.channels

    def __iter__(self) -> Iterable[Tuple[str, Config]]:
        yield from self.channels.items()

    def __getitem__(self, item) -> Config:
        """gets a channels Config, creates if not exists"""
        if item not in self:
            self.load_permissions(item)
        return self.channels[item]


# permissions are requested/generated in channel.py in __init__ for Channel class
perms = Permissions()


def generate_permission_files(*extra_channels):
    from .config import cfg
    for channel in chain(cfg.channels, extra_channels):
        perms.load_permissions(channel, force_update=False)



from datetime import datetime
from typing import Dict, Tuple

from aiohttp import ClientSession, ClientResponse
from async_timeout import timeout

from ..config import get_client_id, get_oauth
from ..data import UserFollowers, UserInfo, RateLimit

__all__ = ('CHANNEL_CHATTERS_URL', 'get_channel_chatters', 'get_stream_data', 'get_url', 'get_user_data', 'get_user_id',
           'STREAM_API_URL', 'USER_API_URL', 'get_user_followers', 'USER_FOLLOWERS_API_URL', 'get_headers',
           'get_user_info', 'get_user_creation_date', 'USER_ACCOUNT_AGE_API')

USER_API_URL = 'https://api.twitch.tv/helix/users?login={}'
STREAM_API_URL = 'https://api.twitch.tv/helix/streams?user_login={}'
CHANNEL_CHATTERS_URL = 'https://tmi.twitch.tv/group/user/{}/chatters'
USER_FOLLOWERS_API_URL = 'https://api.twitch.tv/helix/users/follows?to_id={}'
USER_ACCOUNT_AGE_API = 'https://api.twitch.tv/kraken/users/{}'

user_id_cache: Dict[str, int] = {}


async def get_url(url: str, headers: dict = None) -> Tuple[ClientResponse, dict]:
    headers = headers or get_headers()
    async with ClientSession(headers=headers) as session:
        async with timeout(10):
            async with session.get(url) as resp:
                return resp, await resp.json()


async def get_user_info(user: str) -> UserInfo:
    _, json = await get_url(USER_API_URL.format(user), get_headers())

    if 'error' in json or not json.get('data', None):
        return UserInfo(-1, '', '', '', '', '', '', '', -1)

    data = json['data'][0]
    return UserInfo(
        id=int(data['id']),
        login=data['login'],
        display_name=data['display_name'],
        type=data['type'],
        broadcaster_type=data['broadcaster_type'],
        description=data['description'],
        profile_image_url=data['profile_image_url'],
        offline_image_url=data['offline_image_url'],
        view_count=data['view_count']
    )


async def get_user_creation_date(user: str) -> datetime:
    _, json = await get_url(USER_ACCOUNT_AGE_API.format(user), get_headers(use_kraken=True))

    if 'created_at' not in json:
        return datetime.min()
    #                                            2012-09-03T01:30:56Z
    return datetime.strptime(json['created_at'], '%Y-%m-%dT%H:%M:%SZ')


async def get_user_followers(user: str, headers: dict = None) -> UserFollowers:
    headers = headers or get_headers()
    user_id = await get_user_id(user, headers)
    _, json = await get_url(USER_FOLLOWERS_API_URL.format(user_id), get_headers())
    ratelimit = RateLimit.from_headers(_.headers)

    # covers invalid user id, or some other API error, such as invalid client-id
    if not json or json.get('status', -1) == 400:
        return UserFollowers(-1, '', -1, '', -1, [])

    return UserFollowers(follower_count=json['total'],
                         following=user,
                         following_id=user_id,
                         name=user,
                         id=user_id_cache[user],
                         followers=json['data'])


async def get_user_data(user: str, headers: dict = None) -> dict:
    headers = headers or get_headers()
    _, json = await get_url(USER_API_URL.format(user), headers)

    if not json.get('data'):
        return {}

    return json['data'][0]


async def get_user_id(user: str, headers: dict = None) -> int:
    headers = headers or get_headers()
    # shortcut if the this user's id was already requested
    if user in user_id_cache:
        return user_id_cache[user]

    data = await get_user_data(user, headers)

    if not data:
        return -1

    user_id_cache[user] = data['id']
    return data['id']


async def get_stream_data(user_id: str, headers: dict = None) -> dict:
    headers = headers or get_headers()
    _, json = await get_url(STREAM_API_URL.format(user_id), headers)

    if not json.get('data'):
        return {}

    return json['data'][0]


async def get_channel_chatters(channel: str) -> dict:
    _, data = await get_url(CHANNEL_CHATTERS_URL.format(channel))
    return data


def get_headers(use_kraken: bool = False):
    headers = {'Client-ID': get_client_id()}
    oauth_key = get_oauth().replace("oauth:", "")

    # Check if we have a valid oauth key
    if not use_kraken and not oauth_key:
        headers.update({'Authorization': f'Bearer {oauth_key}'})
    return headers



from twitchbot import (
    Message,
    add_custom_command,
    get_custom_command,
    delete_custom_command,
    custom_command_exist,
    CustomCommand,
    session,
    cfg,
    Command,
    InvalidArgumentsError)

PERMISSION = 'manange_commands'

PREFIX = cfg.prefix
BLACKLISTED_PREFIX_CHARACTERS = './'


def _verify_resp_is_valid(resp: str):
    return resp[0] not in BLACKLISTED_PREFIX_CHARACTERS


@Command('addcmd', permission=PERMISSION, syntax='<name> <response>',
         help='adds a custom command to the database for the this channel.'
              'place holders: %user : the name of the person that triggered the command,'
              '%uptime : the channels live uptime,'
              '%channel : the channels name')
async def cmd_add_custom_command(msg: Message, *args):
    if len(args) < 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_add_custom_command)

    name, resp = args[0], ' '.join(args[1:])
    name = name.lower()

    if not _verify_resp_is_valid(resp):
        raise InvalidArgumentsError(reason='response cannot have . or / as the starting character',
                                    cmd=cmd_add_custom_command)

    if custom_command_exist(msg.channel_name, name):
        raise InvalidArgumentsError(reason='custom command already exist by that name',
                                    cmd=cmd_add_custom_command)

    if add_custom_command(CustomCommand.create(msg.channel_name, name, resp)):
        await msg.reply('successfully added command')
    else:
        await msg.reply('failed to add command')


@Command('updatecmd', permission=PERMISSION, syntax='<name> <response>',
         help="updates a custom command's response message")
async def cmd_update_custom_command(msg: Message, *args):
    if len(args) < 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_update_custom_command)

    name, resp = args[0], ' '.join(args[1:])
    name = name.lower()

    if not _verify_resp_is_valid(resp):
        raise InvalidArgumentsError(reason='response cannot have . or / as the starting character',
                                    cmd=cmd_update_custom_command)

    cmd = get_custom_command(msg.channel_name, name)
    if cmd is None:
        raise InvalidArgumentsError(reason=f'custom command "{name}" does not exist', cmd=cmd_update_custom_command)

    cmd.response = resp
    session.commit()

    await msg.reply(f'successfully updated {cmd.name}')


@Command('delcmd', permission=PERMISSION, syntax='<name>', help='deletes a custom commands')
async def cmd_del_custom_command(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_del_custom_command)

    cmd = get_custom_command(msg.channel_name, args[0].lower())
    if cmd is None:
        raise InvalidArgumentsError(reason=f'no command found for "{args[0]}"', cmd=cmd_del_custom_command)

    if delete_custom_command(msg.channel_name, cmd.name):
        await msg.reply(f'successfully deleted command {cmd.name}')
    else:
        await msg.reply(f'failed to delete command {cmd.name}')


@Command('cmd', syntax='<name>', help='gets a custom commmands response')
async def cmd_get_custom_command(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_get_custom_command)

    cmd = get_custom_command(msg.channel_name, args[0].lower())
    if cmd is None:
        raise InvalidArgumentsError(reason=f'no command found for "{args[0]}"', cmd=cmd_get_custom_command)

    await msg.reply(f'the response for "{cmd.name}" is "{cmd.response}"')



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .command import Command


class InvalidArgumentsError(Exception):
    """
    used to trigger a command's help message in twitch chat when caller gives the command invalid arguments
    """

    # this is a default for if the class is raised without calling its init, ex: raise InvalidArgumentsError
    reason: str = 'invalid arguments'
    cmd: 'Command' = None

    def __init__(self, reason: str = reason, cmd: 'Command' = None):
        super().__init__(reason)
        self.cmd: 'Command' = cmd
        self.reason: str = reason


class BadTwitchAPIResponse(Exception):
    def __init__(self, endpoint, message):
        super().__init__(f'bad response received from endpoint: {endpoint}\nextra details: {message}')



from twitchbot import (
    Message,
    disable_command,
    enable_command,
    is_command_disabled,
    Command,
    InvalidArgumentsError,
    command_exist,
    cfg_disabled_commands,
    channels,
    reload_whitelisted_commands,
)

MANAGE_COMMANDS_PERMISSION = 'manage_commands'


@Command('disablecmd',
         permission=MANAGE_COMMANDS_PERMISSION,
         syntax='<name>',
         help='disables a command for the current channel')
async def cmd_disable_cmd(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_disable_cmd)

    name = args[0].lower()

    if not command_exist(name):
        raise InvalidArgumentsError(reason=f'no command found for "{name}"', cmd=cmd_disable_cmd)

    if is_command_disabled(msg.channel_name, name):
        await msg.reply(f'{name} is already disabled')
        return

    disable_command(msg.channel_name, name)

    await msg.reply(f'disabled command "{name}"')


@Command('enablecmd',
         permission=MANAGE_COMMANDS_PERMISSION,
         syntax='<name>',
         help='enables a command for the current channel')
async def cmd_enable_cmd(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_enable_cmd)

    name = args[0].lower()
    if not command_exist(name):
        raise InvalidArgumentsError(reason=f'no command found for "{name}"',
                                    cmd=cmd_enable_cmd)

    if not is_command_disabled(msg.channel_name, name):
        await msg.reply(f'{name} is not disabled')
        return

    enable_command(msg.channel_name, name)
    await msg.reply(f'enabled command "{name}"')


@Command('reloaddisabled', permission=MANAGE_COMMANDS_PERMISSION, help='reloads disable commands config')
async def cmd_reload_disabled(msg: Message, *args):
    cfg_disabled_commands.load()
    await msg.reply('reloaded disabled commands config')


@Command('disablecmdglobal', permission=MANAGE_COMMANDS_PERMISSION, help='disables a command for all channels the bot is in')
async def cmd_disable_command_global(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_disable_cmd)

    cmd = args[0].lower()

    if not command_exist(cmd):
        raise InvalidArgumentsError(reason=f'no command found for "{cmd}"', cmd=cmd_disable_cmd)

    for channel in channels.values():
        disable_command(channel.name, cmd)

    await msg.reply(f'disabled command "{cmd}" globally')


@Command('enablecmdglobal',
         permission=MANAGE_COMMANDS_PERMISSION,
         syntax='<name>',
         help='enables a command for all channels the bot is in')
async def cmd_enable_command_global(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_enable_cmd)

    cmd = args[0].lower()
    if not command_exist(cmd):
        raise InvalidArgumentsError(reason=f'no command found for "{cmd}"', cmd=cmd_enable_cmd)

    for channel in channels.values():
        enable_command(channel.name, cmd)

    await msg.reply(f'enabled command "{cmd}" globally')


@Command('reloadcmdwhitelist',
         permission=MANAGE_COMMANDS_PERMISSION,
         help='reloads whitelisted commands from the config file')
async def cmd_reload_command_whitelist(msg: Message, *_):
    reload_whitelisted_commands()
    await msg.reply(f'{msg.mention} reloaded command whitelist')



import os
import sys
import traceback
from asyncio import get_event_loop
from importlib import import_module
from inspect import isclass, getfile, getmodulename
from pathlib import Path
from traceback import print_exc
from typing import Dict, Callable, Any

from .channel import Channel
from .command import Command
from .config import cfg
from .disabled_mods import is_mod_disabled
from .enums import Event
from .events import trigger_event, AsyncEventWrapper
from .message import Message
from .shared import get_bot
from .util import temp_syspath, get_py_files, get_file_name

__all__ = ('ensure_mods_folder_exists', 'Mod', 'register_mod', 'trigger_mod_event', 'mods',
           'load_mods_from_directory', 'mod_exists', 'reload_mod', 'is_mod', 'unregister_mod',
           'ensure_commands_folder_exists')


# noinspection PyMethodMayBeStatic
class Mod:
    name = 'DEFAULT'

    async def loaded(self):
        """
        called when the Mod is loaded and created/registered
        """

    async def unloaded(self):
        """
        called when the Mod is unloaded (aka unregistered)
        """

    # region events
    async def on_enable(self, channel: str):
        """
        triggered when the mod is enabled
        :param channel: the channel the mod is enabled in
        """

    async def on_disable(self, channel: str):
        """
        triggered when the mod is disabled
        :param channel: the channel the mod is disabled in
        """

    async def on_mod_reloaded(self, mod: 'Mod'):
        """
        triggered when a mod is reloaded using reload_mod() or !reloadmod
        :param mod: mod being reloaded
        """

    async def on_bot_banned_from_channel(self, msg: Message, channel: Channel):
        """
        triggered when the bot attempts to join a banned channel
        :param msg: the message that twitch sent saying the bot was banned
        :param channel: the channel the bot was banned from
        """

    async def on_bot_timed_out_from_channel(self, msg: Message, channel: Channel, seconds: int):
        """
        triggered when the bot is timed out on a channel
        :param msg: the message that twitch sent saying the bot was timed out
        :param channel: the channel the bot was timed out on
        :param seconds: how many seconds left in the timeout
        """

    async def on_connected(self):
        """
        triggered when the bot connects to all the channels specified in the config file
        """

    async def on_raw_message(self, msg: Message):
        """
        triggered the instant a message is received,
        this message can be any message received,
        including twitches messages that do not have any useful information
        """

    async def on_privmsg_sent(self, msg: str, channel: str, sender: str):
        """
        triggered when the bot sends a privmsg
        """

    async def on_privmsg_received(self, msg: Message):
        """
        triggered when a privmsg is received, is not triggered if the msg is a command
        """

    async def on_whisper_sent(self, msg: str, receiver: str, sender: str):
        """
        triggered when the bot sends a whisper to someone
        """

    async def on_whisper_received(self, msg: Message):
        """
        triggered when a user sends the bot a whisper
        """

    async def on_permission_check(self, msg: Message, cmd: Command) -> bool:
        """
        triggered when a command permission check is requested
        :param msg: the message the command was found from
        :param cmd: the command that was found
        :return: bool indicating if the user has permission to call the command, True = yes, False = no
        """
        return True

    async def on_before_command_execute(self, msg: Message, cmd: Command) -> bool:
        """
        triggered before a command is executed
        :return bool, if return value is False, then the command will not be executed
        """
        return True

    async def on_after_command_execute(self, msg: Message, cmd: Command):
        """
        triggered after a command has executed
        """

    async def on_bits_donated(self, msg: Message, bits: int):
        """
        triggered when a bit donation is posted in chat
        """

    async def on_channel_raided(self, channel: Channel, raider: str, viewer_count: int):
        """
        triggered when the channel is raided
        :param channel: the channel who was raided
        :param raider: the user who raided
        :param viewer_count: the number of viewers who joined in the raid
        """

    async def on_channel_joined(self, channel: Channel):
        """
        triggered when the bot joins a channel
        """

    async def on_user_join(self, user: str, channel: Channel):
        """
        triggered when a user joins a channel the bot is in
        :param user: the user who joined
        :param channel: the channel that the user joined
        """

    async def on_user_part(self, user: str, channel: Channel):
        """
        triggered when a user leaves from a channel the bot is in
        :param user: the user who left
        :param channel: the channel that the user left
        """

    async def on_channel_subscription(self, subscriber: str, channel: Channel, msg: Message):
        """
        triggered when a user subscribes
        """

    async def on_channel_points_redemption(self, msg: Message, reward: str):
        """
        triggered when a viewers redeems channel points for a reward
        """

    # endregion


mods: Dict[str, Mod] = {}


def register_mod(mod: Mod) -> bool:
    """
    registers a mod globally
    :param mod: the mod to register
    :return: if registration was successful
    """
    if mod.name in mods:
        return False

    mods[mod.name] = mod
    get_event_loop().create_task(mod.loaded())
    return True


def unregister_mod(mod: Mod) -> bool:
    """
    unregisters a mod from the global cache `mods`
    :param mod: mod to unregister
    :return: if it successfully unregistered it
    """
    if mod.name not in mods:
        return False

    get_event_loop().create_task(mod.unloaded())
    del mods[mod.name]
    return True


async def trigger_mod_event(event: Event, *args, channel: str = None) -> list:
    """
    triggers a event on all mods
    if the channel is passed, the it is checked if the mod is enabled for that channel,
    if not, the event for that mod is skipped
    :param event: the event to raise on all the mods
    :param args: the args to pass to the event
    :param channel: the channel the event is being raised from
    :return: the result of all the mod event calls in a list
    """

    async def _missing_function(*ignored):
        pass

    output = []
    for mod in mods.values():
        if channel is not None and is_mod_disabled(channel, mod.name):
            continue

        try:
            output.append(await getattr(mod, event.value, _missing_function)(*args))
        except Exception as e:
            print(f'\nerror has occurred while triggering a event on a mod, details:\n'
                  f'mod: {mod.name}\n'
                  f'event: {event}\n'
                  f'error: {type(e)}\n'
                  f'reason: {e}\n'
                  f'stack trace:')
            traceback.print_exc()
    return output


def ensure_mods_folder_exists():
    """
    creates the mod folder if it does not exists
    """
    if not os.path.exists(cfg.mods_folder):
        os.mkdir(cfg.mods_folder)


def ensure_commands_folder_exists():
    """
    creates the commands folder if it does not exists
    """
    if not os.path.exists(cfg.commands_folder):
        os.mkdir(cfg.commands_folder)


def load_mods_from_directory(fullpath, predicate: Callable[[str, Any], bool] = None, log=True):
    """
    loads all mods from the given directory, only .py files are loaded
    :param fullpath: the path to search for mods to load
    """
    if log:
        print('loading mods from:', fullpath)

    with temp_syspath(fullpath):
        for file in get_py_files(fullpath):
            # we need to import the module to get its attributes
            module = import_module(get_file_name(file))
            for obj in module.__dict__.values():
                # verify the obj is a class, is a subclass of Mod, and is not Mod class itself
                if not is_mod(obj):
                    continue
                # create a instance of the mod subclass, then register it
                if predicate is None or (predicate is not None and predicate(obj.name, obj)):
                    register_mod(obj())


def reload_mod(mod_name: str):
    mod = mods.get(mod_name)
    if mod is None:
        raise ValueError(f'could not find mod by the name of "{mod_name}"')

    try:
        # get the module's file path, this is needed to add it to python's import search paths
        path = Path(getfile(mod.__class__))
        with temp_syspath(path.parent):
            # this is needed to make python import the latest version from disk,
            # python caches imports in sys.modules
            # doing this removes it from the cache, so the latest version from the disk is imported
            prev_module = sys.modules[getmodulename(path)]
            del sys.modules[getmodulename(path)]
            # this dict stores the Mod's / global objects in the Mod's .py file
            # it lets us iterate over its value to check for the Mod that we want to reload
            module = __import__(getmodulename(path), locals={}, globals={})
            for item in module.__dict__.values():
                if is_mod(item) and item.name == mod.name:
                    unregister_mod(mod)
                    reloaded_mod = item()
                    register_mod(reloaded_mod)

                    # removed previous events that were registered via @event_handler
                    # this prevents event duplication and stacking
                    for var in prev_module.__dict__.values():
                        if isinstance(var, AsyncEventWrapper):
                            var.unregister()

                    # trigger events
                    get_event_loop().create_task(trigger_mod_event(Event.on_mod_reloaded, reloaded_mod))
                    get_event_loop().create_task(get_bot().on_mod_reloaded(reloaded_mod))
                    get_event_loop().create_task(trigger_event(Event.on_mod_reloaded, reloaded_mod))
                    return True

    except Exception as e:
        print(f'error trying to reload Mod "{mod.name}", error type: {type(e)}, error: {e}')
        print_exc()
    return False


def is_mod(obj):
    return isclass(obj) and issubclass(obj, Mod) and obj is not Mod


def mod_exists(mod: str) -> bool:
    """
    returns of a mod exists
    :param mod: the mod to check for
    :return: bool indicating if the mod exists
    """
    return mod in mods



from enum import Enum, IntFlag, auto

__all__ = ('Event', 'CommandContext', 'MessageType', 'UserType')


class NamedEnum(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self


class UserType(NamedEnum):
    GLOBAL_MOD = auto()
    ADMIN = auto()
    STAFF = auto()
    AFFILIATE = auto()
    NORMAL = auto()


class MessageType(NamedEnum):
    PRIVMSG = auto()
    WHISPER = auto()
    COMMAND = auto()
    PING = auto()
    USER_JOIN = auto()
    USER_PART = auto()
    SUBSCRIPTION = auto()
    RAID = auto()
    USER_NOTICE = auto()
    CHANNEL_POINTS_REDEMPTION = auto()
    BITS = auto()
    NOTICE = auto()
    BOT_PERMANENTLY_BANNED = auto()
    BOT_TIMED_OUT = auto()
    NONE = auto()


class CommandContext(IntFlag):
    CHANNEL = auto()
    WHISPER = auto()
    BOTH = CHANNEL | WHISPER


class Event(NamedEnum):
    on_before_command_execute = auto()
    on_after_command_execute = auto()
    on_bits_donated = auto()
    on_channel_subscription = auto()
    on_channel_raided = auto()
    on_channel_joined = auto()
    on_connected = auto()
    on_privmsg_received = auto()
    on_privmsg_sent = auto()
    on_whisper_received = auto()
    on_whisper_sent = auto()
    on_permission_check = auto()
    on_raw_message = auto()
    on_user_join = auto()
    on_user_part = auto()
    on_mod_reloaded = auto()
    on_channel_points_redemption = auto()
    on_bot_banned_from_channel = auto()
    on_bot_timed_out_from_channel = auto()



import webbrowser
from collections import namedtuple

__all__ = ('show_auth_gui', 'AuthInfo')

OAUTH_GENERATOR_URL = 'https://twitchapps.com/tmi/'
AuthInfo = namedtuple('AuthInfo', 'username oauth')


def show_auth_gui():
    try:
        import tkinter as tk
    except ImportError:
        print('unable to import tkinter to show the GUI!')
        return

    root = tk.Tk()
    width = 30

    tk.Label(root, text='bot account name:').grid(row=0, column=0)
    username = tk.StringVar(root)
    tk.Entry(root, textvar=username, width=width).grid(row=0, column=1)

    tk.Label(root, text='oauth:').grid(row=1, column=0)
    password = tk.StringVar(root)
    tk.Entry(root, show='*', textvar=password, width=width).grid(row=1, column=1)

    tk.Button(
        text='open twitch oauth generator website',
        command=lambda: webbrowser.open(OAUTH_GENERATOR_URL)
    ).grid(row=2, column=0)

    tk.Button(root, text='im done', command=lambda: root.destroy()).grid(row=2, column=1)

    root.mainloop()
    return AuthInfo(username=username.get(), oauth=password.get())



from dataclasses import dataclass

from ..exceptions import BadTwitchAPIResponse
from ..util import get_channel_chatters, CHANNEL_CHATTERS_URL

__all__ = [
    'Chatters',
    'CHANNEL_CHATTERS_URL'
]

VIPS = 'vips'
MODERATORS = 'moderators'
STAFF = 'staff'
ADMINS = 'admins'
GLOBAL_MODS = 'global_mods'
VIEWERS = 'viewers'
CHATTER_COUNT = 'chatter_count'
CHATTERS = 'chatters'


@dataclass
class Chatters:
    channel: str
    mods: frozenset = frozenset()
    vips: frozenset = frozenset()
    staff: frozenset = frozenset()
    admins: frozenset = frozenset()
    global_mods: frozenset = frozenset()
    viewers: frozenset = frozenset()
    all_viewers: frozenset = frozenset()
    viewer_count: int = 0

    async def update(self):
        json = ''
        try:
            json = await get_channel_chatters(self.channel)
            self._verify_base_response_is_valid(json)
            chatters = json['chatters']
            self._verify_chatters_response_is_valid(chatters)

            self.mods = frozenset(chatters[MODERATORS])
            self.vips = frozenset(chatters[VIPS])
            self.staff = frozenset(chatters[STAFF])
            self.admins = frozenset(chatters[ADMINS])
            self.global_mods = frozenset(chatters[GLOBAL_MODS])
            self.viewers = frozenset(chatters[VIEWERS])
            self.viewer_count = json[CHATTER_COUNT]
            self.all_viewers = self.vips | self.mods | self.staff | self.admins | self.global_mods | self.viewers | {self.channel}
        except Exception as e:
            print(f'\nCHATTERS API ERROR\njson received: {json}\n{e}\nEND CHATTERS API ERROR\n')

    def __contains__(self, item):
        return item.lower() in self.all_viewers

    def __iter__(self):
        yield from self.all_viewers

    def _verify_response_is_dict(self, json):
        if not isinstance(json, dict):
            raise BadTwitchAPIResponse(CHANNEL_CHATTERS_URL,
                                       f'response was not the expected type, expected: {dict}, actual: {type(json)}'
                                       f'\nvalue: {json}')

    def _verify_keys(self, json, keys):
        for key in keys:
            if key not in json:
                raise BadTwitchAPIResponse(CHANNEL_CHATTERS_URL, f'response is missing required key: {key}')

    def _verify_base_response_is_valid(self, json):
        self._verify_response_is_dict(json)
        self._verify_keys(json, (CHATTER_COUNT, CHATTERS))

    def _verify_chatters_response_is_valid(self, json):
        self._verify_response_is_dict(json)
        self._verify_keys(json, (VIPS, MODERATORS, STAFF, ADMINS, GLOBAL_MODS, VIEWERS))



from collections import namedtuple
from datetime import datetime, timedelta
from secrets import choice
from typing import Dict, Optional, Tuple

__all__ = (
    'accept_duel',
    'duel_exists',
    'DUEL_DURATION',
    'add_duel',
    'remove_duel',
    'duel_expired',
    'Duel',
    'get_duel'
)

Duel = namedtuple('Duel', 'challenger target bet time')
DUEL_DURATION = timedelta(seconds=30)

# keys are: (channel, challenger, target)
# keeps track of all pending duels
pending_duels: Dict[Tuple[str, str, str], Duel] = {}


def _key(channel: str, challenger: str, target: str):
    return channel, challenger.lower(), target.lower()


def duel_expired(duel: Duel) -> bool:
    return datetime.now() - duel.time > DUEL_DURATION


def duel_exists(channel: str, challenger: str, target: str) -> bool:
    """
    returns True if a duel exist, else False

    :param channel: the channel which the challenger / target is in
    :param challenger: the user who initiated the duel
    :param target: the user who is being challenged
    :return: True if a duel exist between challenger and target, else False
    """
    return _key(channel, challenger, target) in pending_duels


def add_duel(channel: str, challenger: str, target: str, bet: int):
    """
    adds a new pending duel

    :param channel: the channel which the challenger / target is in
    :param challenger: the user who initiated the duel
    :param target: the user who is being challenged
    :param bet: the bet that the winner takes, and the loser loses
    """
    key = _key(channel, challenger, target)
    pending_duels[key] = Duel(challenger.lower(), target.lower(), bet, datetime.now())


def remove_duel(channel: str, challenger: str, target: str):
    key = _key(channel, challenger, target)
    if key in pending_duels:
        del pending_duels[key]


def get_duel(channel: str, challenger: str, target: str) -> Optional[Duel]:
    """
    gets a pending duel, if duel does not exist, return None

    :param channel: the channel which the challenger / target is in
    :param challenger: the user who initialized the duel
    :param target: the user who was challenged
    :return: duel if it exists, else None
    """
    return pending_duels.get(_key(channel, challenger, target))


def accept_duel(channel: str, challenger: str, target: str) -> Tuple[Optional[str], Optional[int]]:
    """
    accepts a duel from [target] issued by [challenger]
    returns tuple of (winner, bet) if duel exists and has not expired, else (None, None)

    :param channel: the channel which the challenger / target is in
    :param challenger: the user who initialized a challenge to [target]
    :param target: the user who was challenged
    :return: (winner, bet) if duel exists and has not expired, else (None, None)
    """
    duel = get_duel(channel, challenger, target)
    remove_duel(channel, challenger, target)

    if not duel or duel_expired(duel):
        return None, None

    return choice((challenger, target)), duel.bet



from socket import socket

HOST = input('enter command server host (leave blank for "localhost"): ').strip() or 'localhost'
PORT = int(input('enter command server port (leave blank for 1337): ').strip() or 1337)

s = socket()
s.connect((HOST, PORT))
SIZE = 300


def read():
    return s.recv(SIZE).decode('utf8')


def send(text):
    s.send(f'{text}\n'.encode())


# prints the welcome message
print(read())

# wait for a valid channel to be selected
msg = ''
channel = ''

while True:
    msg = read()
    if 'quit' in msg:
        break

    print(msg)
    channel = input('>>> ')
    send(channel)

# channel was selected, now start the command/message send loop
print('type `exit` to close this window')

while True:
    text = input(f'({channel}): ')
    if text.lower().strip() == 'exit':
        break

    send(text)



from asyncio import Task

from sqlalchemy import Column, Integer, String, Float, Boolean

from .session import Base, init_tables
from ..config import cfg
from ..enums import CommandContext

__all__ = ('Quote', 'CustomCommand', 'Balance', 'CurrencyName', 'MessageTimer')


class Quote(Base):
    __tablename__ = 'quotes'

    id = Column(Integer, primary_key=True, nullable=False)
    user = Column(String(255))
    channel = Column(String(255), nullable=False)
    alias = Column(String(255))
    value = Column(String(255), nullable=False)

    @classmethod
    def create(cls, channel: str, value: str, user: str = None, alias: str = None):
        return Quote(channel=channel.lower(), user=user, value=value, alias=alias)


class CustomCommand(Base):
    __tablename__ = 'commands'

    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String(255), nullable=False)
    channel = Column(String(255), nullable=False)
    response = Column(String(255), nullable=False)
    context = CommandContext.CHANNEL
    permission = None

    @classmethod
    def create(cls, channel: str, name: str, response: str):
        return CustomCommand(channel=channel.lower(), name=name.lower(), response=response)

    @property
    def fullname(self):
        return self.name

    def __str__(self):
        return f'<CustomCommand channel={self.channel!r} name={self.name!r} response={self.response!r}>'


class Balance(Base):
    __tablename__ = 'balance'

    id = Column(Integer, nullable=False, primary_key=True)
    channel = Column(String(255), nullable=False)
    user = Column(String(255), nullable=False)
    balance = Column(Integer, nullable=False)

    @classmethod
    def create(cls, channel: str, user: str, balance: int = cfg.default_balance):
        return Balance(channel=channel.lower(), user=user, balance=balance)


class CurrencyName(Base):
    __tablename__ = 'currency_names'

    id = Column(Integer, nullable=False, primary_key=True)
    channel = Column(String(255), nullable=False)
    name = Column(String(255), nullable=False)

    @classmethod
    def create(cls, channel: str, name: str):
        return CurrencyName(channel=channel.lower(), name=name)


class MessageTimer(Base):
    __tablename__ = 'message_timers'

    id = Column(Integer, nullable=False, primary_key=True)
    name = Column(String(255), nullable=False)
    channel = Column(String(255), nullable=False)
    message = Column(String(255), nullable=False)
    interval = Column(Float, nullable=False)
    active = Column(Boolean, nullable=False, default=False)
    task: Task = None

    @property
    def running(self):
        return self.task is not None and not self.task.done()

    @classmethod
    def create(cls, channel: str, name: str, message: str, interval: float, active=False):
        return MessageTimer(name=name, channel=channel, message=message, interval=interval, active=active)


init_tables()



from .baseapi import Api
from ..enums import UserType
from .. import util


class UserInfoApi(Api):
    def __init__(self, client_id: str, user: str):
        super().__init__(client_id, user)

        self.id = 0
        self.login = ''
        self.display_name = ''
        self.type = UserType.NORMAL
        self.description = ''
        self.profile_image_url = ''
        self.offline_image_url = ''
        self.view_count = ''

    async def update(self, log=False):
        data = await util.get_user_data(self.user, self.headers)

        try:
            self.id = data['id']
            self.login = data['login']
            self.display_name = data['display_name']
            self.type = UserType(data['type'])
            self.description = data['description']
            self.profile_image_url = data['profile_image_url']
            self.offline_image_url = data['offline_image_url']
            self.view_count = data['view_count']

            await self.on_successful_update()
        except KeyError:
            if log:
                print(f'[USER INFO API] failed to update data for {self.user}')

            await self.on_failed_update()



import re

RE_PRIVMSG = re.compile(
    r'(?P<tags>.*):'
    r'(?P<user>[\w\d]+)!(?P=user)@(?P=user)\.tmi\.twitch\.tv PRIVMSG #(?P<channel>[\w\d]+) :(?P<content>.+)'
)

# example whisper
# :nickname!nickname@nickname.tmi.twitch.tv WHISPER bob :hello world!
RE_WHISPER = re.compile(
    r':(?P<user>[\w\d]+)!(?P=user)@(?P=user)\.tmi\.twitch\.tv WHISPER (?P<receiver>[\w\d]+) :(?P<content>.+)'
)

# example joining channel
# :nickname!nickname@nickname.tmi.twitch.tv JOIN #bob
RE_USER_JOIN = re.compile(
    r':(?P<user>[\w\d]+)!(?P=user)@(?P=user)\.tmi\.twitch\.tv JOIN #(?P<channel>\w+)'
)

# user leaves the channel
#  :nickname!nickname@nickname.tmi.twitch.tv PART #bob
RE_USER_PART = re.compile(
    r':(?P<user>[\w\d]+)!(?P=user)@(?P=user)\.tmi\.twitch\.tv PART #(?P<channel>\w+)'
)

# finds mentions in twitch messages
# example: hello @bob!
RE_AT_MENTION = re.compile(
    r'@([\w\d]+)'
)

# user notices / subscriptions
RE_USERNOTICE = re.compile(
    r'(?P<tags>.*):tmi\.twitch\.tv USERNOTICE #(?P<channel>[\w\d]+)(?: :)?(?P<content>.+)?'
)

# user notices / subscriptions
# example:
# @msg-id=msg_banned :tmi.twitch.tv NOTICE #X :You are permanently banned from talking in X.
RE_NOTICE = re.compile(
    r'(?P<tags>.*):tmi\.twitch\.tv NOTICE #(?P<channel>[\w\d]+)(?: :)?(?P<content>.+)?'
)

#  @msg-id=msg_timedout :tmi.twitch.tv NOTICE #X :You are timed out for 99906 more seconds.
RE_TIMEOUT_DURATION = re.compile(
    r'timed out for (?P<seconds>\d+)'
)



import os
import json
from pathlib import Path
from typing import Optional

from .gui import show_auth_gui

__all__ = ('cfg', 'Config', 'mysql_cfg', 'CONFIG_FOLDER', 'generate_config', 'get_oauth', 'get_nick', 'get_client_id',
           'DEFAULT_NICK', 'DEFAULT_OAUTH', 'DEFAULT_CLIENT_ID', 'is_config_valid')

CONFIG_FOLDER = Path('configs')


# noinspection PyTypeChecker
class Config:
    def __init__(self, file_path: Path, **defaults):
        self.file_path = file_path
        self.data = {}
        self.defaults = defaults

        self.load()
        self._validate()

    def _validate(self):
        """checks that all default keys are present"""
        for k, v in self.defaults.items():
            if k not in self.data:
                self.data[k] = v

        self.save()

    @property
    def exist(self):
        """returns if the config file exist"""
        return self.file_path.exists()

    @property
    def in_folder(self):
        """returns if the config file is in a folder"""
        return len(os.path.split(self.file_path)) > 1

    def regen(self):
        """restores config file to its default state and values"""
        self.create(overwrite=True)
        self.load()

    def save(self):
        """updates the config file with the current config data"""
        with open(self.file_path, 'w') as file:
            json.dump(self.data, file, indent=2)

    def load(self):
        """
        loads the config file's contents into this config object's `data` attribute
        creates the config if it doesnt exist
        """
        if not self.exist:
            self.create()

        with open(self.file_path) as file:
            self.data = json.load(file)

    def create(self, overwrite=False):
        """creates the config json file, does nothing if it already exist, except if `ignore_exist` is True"""
        if self.exist and not overwrite:
            return

        if self.in_folder:
            try:
                os.makedirs(os.path.join(os.curdir, *os.path.split(self.file_path)[:-1]))
            except FileExistsError:
                pass

        with open(self.file_path, 'w') as file:
            json.dump(self.defaults, file, indent=2)

    def prompt_edit_oauth(self):
        pass

    def __getattr__(self, item):
        """allows for getting config values by accessing a attribute"""
        return self.__dict__[item] if item in self.__dict__ else self.data.get(item)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.data[key] = value
        self.save()

    def __contains__(self, item):
        return item in self.data

    def __iter__(self):
        yield from self.data.items()


DEFAULT_OAUTH = 'oauth:'
DEFAULT_NICK = 'nick'
DEFAULT_CLIENT_ID = 'CLIENT_ID'

cfg = Config(
    file_path=CONFIG_FOLDER / 'config.json',
    nick=DEFAULT_NICK,
    oauth=DEFAULT_OAUTH,
    client_id=DEFAULT_CLIENT_ID,
    prefix='!',
    default_balance=200,
    loyalty_interval=60,
    loyalty_amount=2,
    owner='BOT_OWNER_NAME',
    channels=['channel'],
    mods_folder='mods',
    commands_folder='commands',
    command_server_enabled=True,
    command_server_port=1337,
    command_server_host='localhost',
    disable_whispers=False,
    use_command_whitelist=False,
    send_message_on_command_whitelist_deny=True,
    command_whitelist=[
        'help', 'commands', 'reloadcmdwhitelist', 'reloadmod', 'reloadperms', 'disablemod', 'enablemod', 'disablecmdglobal', 'disablecmd',
        'enablecmdglobal', 'enablecmd', 'addcmd', 'delcmd', 'updatecmd', 'cmd'
    ],
)

mysql_cfg = Config(
    CONFIG_FOLDER / 'mysql.json',
    enabled=False,
    address='localhost',
    port='3306',
    username='root',
    password='password',
    database='twitchbot',
)


def generate_config():
    if not is_config_valid(check_client_id=False):
        if input('show the bot config GUI? [Y/N]: ').lower() == 'y':
            print('please enter your twitch auth info the the GUI that will pop-up shortly')
            auth = show_auth_gui()
            cfg['nick'] = auth.username
            cfg['oauth'] = auth.oauth
        else:
            print(
                'open the config folder at <BOT_FOLDER>/configs/config.json\nthen enter the bot oauth/name, the owner, and the channels to join')
            input('\npress enter to close bot...')
            return False

    if 'oauth:' not in cfg['oauth']:
        print('oauth must start with `oauth:` and be followed by the token itself, ex: oauth:exampletoken12')
        input('press enter to exit...')
        return False

    if len(cfg['oauth']) <= 10:
        print('oauth is too short, must be `oauth:` followed by the token itself, ex: oauth:exampletoken12')
        input('press enter to exit...')
        return False

    # this is to fix the edge case where the user entered enters cased names / channel names
    # without this, the twitch API returns weird responses that are not easy to figure out why it is doing it
    cfg['nick'] = cfg['nick'].lower()
    cfg['owner'] = cfg['owner'].lower()
    cfg['channels'] = [chan.lower() for chan in cfg['channels']]
    cfg.save()

    return True


def is_config_valid(check_client_id=False):
    if (get_nick() == DEFAULT_NICK or get_oauth() == DEFAULT_OAUTH
            or (check_client_id and get_client_id() == DEFAULT_CLIENT_ID)):
        return False
    return True


def get_nick() -> str:
    """
    gets the bot accounts NICK, this is the login username for the bot account

    if the the NICK matches the pattern `ENV_KEY_HERE` it will get `KEY_HERE` from os.environ, else it just grabs it from the cfg
    """
    nick = cfg.nick
    if _is_env_key(nick):
        value = _get_env_value(nick)
        if value is None:
            print(f'could not get NICK from environment with key: {nick[4:]}')
            input('\npress enter to exit...')
            exit(1)
        return value
    return nick


def get_oauth() -> str:
    """
    gets the bot accounts OAUTH
    this is needed because of supporting getting it from os.environ

    if the the OAUTH matches the pattern `ENV_KEY_HERE` it will get `KEY_HERE` from os.environ, else it just grabs it from the cfg
    """
    oauth = cfg.oauth
    if _is_env_key(oauth):
        value = _get_env_value(oauth)
        if value is None:
            print(f'could not get OAUTH from environment with key: {oauth[4:]}')
            input('\npress enter to exit...')
            exit(1)
        return value
    return oauth


def get_client_id() -> str:
    """
    gets the bot accounts CLIENT_ID
    this is needed because of supporting getting it from os.environ

    if the the CLIENT_ID matches the pattern `ENV_KEY_HERE` it will get `KEY_HERE` from os.environ, else it just grabs it from the cfg
    """
    client_id = cfg.client_id
    if _is_env_key(client_id):
        value = _get_env_value(client_id)
        if value is None:
            print(f'could not get CLIENT_ID from environment with key: {client_id[4:]}')
            input('\npress enter to exit...')
            exit(1)
        return value
    return client_id


def _is_env_key(key) -> bool:
    return key.lower().startswith('env_')


def _get_env_value(key) -> Optional[str]:
    if key.lower().startswith('env_'):
        key = key[4:]

    return os.environ.get(key)



from twitchbot import BaseBot, Mod, Event


def test_base_bot_events_are_set():
    for event in Event:
        assert event.name in BaseBot.__dict__, f'BaseBot must implement event {event}'


def test_mod_events_are_set():
    for event in Event:
        assert event.name in Mod.__dict__, f'Mod must implement event {event}'



from twitchbot import Mod, Message, reply_wait_queue, MessageType


class ReplyWaiter(Mod):
    name = 'replywaiter'

    async def on_raw_message(self, msg: Message):
        if not reply_wait_queue or msg.type not in {MessageType.WHISPER, MessageType.PRIVMSG}:
            return

        remove = []
        for i, (future, predicate) in enumerate(reply_wait_queue):
            if await predicate(msg):
                # set the result of the future to the message, this makes the future complete and return its result
                future.set_result(msg)
                remove.append(i)

        for i in reversed(remove):
            del reply_wait_queue[i]



from .arena import *
from .channel import *
from twitchbot.api.chatters import *
from .colors import *
from .command import *
from .config import *
from .enums import *
from .irc import *
from .message import *
from .permission import *
from .ratelimit import *
from .regex import *
from .util import *
from .database import *
from .bots import *
from .api import *
from .events import *
from .loyalty_ticker import *
from .disabled_commands import *
from .duel import *
from .command_server import start_command_server
from .modloader import *
from .disabled_mods import *
from .data import *
from .exceptions import *
from .shared import *
from .replywaiter import *
from .command_whitelist import *
from . import builtin_commands
from . import builtin_mods

import os

load_commands_from_directory(os.path.join(__path__[0], 'builtin_commands'))

load_mods_from_directory(os.path.join(__path__[0], 'builtin_mods'))
load_mods_from_directory(os.path.abspath(cfg.mods_folder))

ensure_mods_folder_exists()
ensure_commands_folder_exists()

load_commands_from_directory(os.path.abspath(cfg.commands_folder))
start_command_server()



from operator import attrgetter

from twitchbot import (
    Command,
    commands,
    CommandContext,
    Message, get_all_custom_commands,
    InvalidArgumentsError,
    get_command)


# @Command('list')
# async def cmd_list(msg: Message, *args):
#     for c in channels.values():
#         await msg.reply(
#             whisper=True,
#             msg=f'channel: {c.name}, viewers: {c.chatters.viewer_count}, is_mod: {c.is_mod}, is_live: {c.live}')


@Command('commands', context=CommandContext.BOTH, help='lists all commands')
async def cmd_commands(msg: Message, *args):
    custom_commands = ', '.join(map(attrgetter('name'), get_all_custom_commands(msg.channel_name)))
    global_commands = ', '.join(map(attrgetter('fullname'), commands.values()))

    await msg.reply(whisper=True, msg=f'GLOBAL: {global_commands}')
    if custom_commands:
        await msg.reply(whisper=True, msg=f'CUSTOM: {custom_commands}')


@Command(name='help', syntax='<command>', help='gets the help text for a command')
async def cmd_help(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_help)

    cmd = get_command(args[0])
    if not cmd:
        raise InvalidArgumentsError(reason=f'command not found', cmd=cmd_help)

    await msg.reply(msg=f'help for {cmd.fullname} - syntax: {cmd.syntax} - help: {cmd.help}')



import asyncio
from .. import util

from datetime import datetime


class Api:
    def __init__(self, client_id: str, user: str):
        self.user = user
        self.client_id = client_id
        self.headers = {'Client-ID': client_id}

    async def update(self, log=False):
        pass

    async def update_loop(self, delay=60):
        while True:
            await self.update()
            await asyncio.sleep(delay)

    async def on_successful_update(self):
        """
        called when the API updates without any errors
        """
        pass

    async def on_failed_update(self):
        """
        called when the API encounters a error trying to update
        """
        pass

    def __eq__(self, other):
        if not isinstance(other, Api):
            return False

        return self.user == other.user



from typing import List

from .config import cfg

__all__ = [
    'command_whitelist_enabled',
    'whitelisted_commands',
    'remove_command_from_whitelist',
    'add_command_to_whitelist',
    'is_command_whitelisted',
    'reload_whitelisted_commands',
    'send_message_on_command_whitelist_deny',
]


def send_message_on_command_whitelist_deny():
    return cfg.send_message_on_command_whitelist_deny


def command_whitelist_enabled() -> bool:
    return cfg.use_command_whitelist


def whitelisted_commands() -> List[str]:
    return cfg.command_whitelist


def add_command_to_whitelist(cmd_name: str, save: bool = True):
    cmd_name = cmd_name.lower()
    cmds = whitelisted_commands()
    if cmd_name not in cmds:
        cmds.append(cmd_name)
        if save:
            cfg.save()


def remove_command_from_whitelist(cmd_name: str, save: bool = True):
    cmd_name = cmd_name.lower()
    cmds = whitelisted_commands()
    if cmd_name in cmds:
        cmds.remove(cmd_name)
        if save:
            cfg.save()


def is_command_whitelisted(cmd_name: str):
    if not command_whitelist_enabled():
        return True
    return cmd_name.lower() in whitelisted_commands()


def reload_whitelisted_commands():
    cfg.load()



from typing import NamedTuple

__all__ = ['UserInfo']


class UserInfo(NamedTuple):
    id: int
    login: str
    display_name: str
    type: str
    broadcaster_type: str
    description: str
    profile_image_url: str
    offline_image_url: str
    view_count: int



from asyncio import ensure_future, Task, sleep, get_event_loop, Future
from typing import Dict, Coroutine, Optional, Tuple

__all__ = ('active_tasks', 'add_task', 'get_task', 'task_running', 'stop_task', 'task_exist', 'stop_all_tasks', 'add_nameless_task')

active_tasks: Dict[str, Task] = {}


def add_task(name: str, coro: Coroutine):
    active_tasks[name.lower()] = ensure_future(coro)


nameless_task_counter = 0


def add_nameless_task(coro: Coroutine) -> Tuple[str, Future]:
    global nameless_task_counter
    nameless_task_name = f'nameless_task_{nameless_task_counter}'
    nameless_task_counter += 1
    future = ensure_future(coro)
    active_tasks[nameless_task_name] = future
    return nameless_task_name, future


def get_task(name: str) -> Optional[Task]:
    return active_tasks.get(name.lower())


def task_exist(name: str):
    return name.lower() in active_tasks


def stop_all_tasks():
    for task in active_tasks.values():
        task.cancel()


def stop_task(name: str) -> bool:
    """stops a task, returns if it was successful"""
    name = name.lower()

    if not task_exist(name) or not task_running(name):
        return False

    task = get_task(name)
    task.cancel()

    del active_tasks[name]


def task_running(name: str):
    return task_exist(name) and not get_task(name).done()


async def _remove_done_tasks():
    global active_tasks

    while True:
        done_tasks = tuple(k for k, t in active_tasks.items() if t.done())
        for k in done_tasks:
            del active_tasks[k]

        await sleep(15)


get_event_loop().create_task(_remove_done_tasks())



from twitchbot import Message, Command, CommandContext

pings = 0


@Command('ping', cooldown=3)
async def cmd_ping(msg: Message, *args):
    global pings
    pings += 1
    await msg.reply(f'Pong #{pings}')

# @Command('whisper', context=CommandContext.WHISPER)
# async def cmd_whisper(msg: Message, *args):
#     await msg.reply('got it!')

# @Command('shoutout', permission='shoutout')
# async def cmd_shoutout(msg: Message, *args):
#     info = await get_user_info(args[0])
#     date = format_datetime(await get_user_creation_date(args[0]))
#     followers = await get_user_followers(args[0])
#
#     await msg.reply(f'go give {args[0]} a follow at {f"https://twitch.tv/{args[0]}"}')
#     await msg.reply(
#         f'{args[0]} with {followers.follower_count} followers and {info.view_count} views * created at {date}')

# @Command('test')
# async def cmd_test(msg: Message, *args):
#     await msg.reply(f'{format_datetime(await get_user_creation_date(msg.author))}')



import traceback
from collections import defaultdict
from typing import DefaultDict, Callable, List

from .enums import Event

custom_event_handlers: DefaultDict[Event, List[Callable]] = defaultdict(list)


async def trigger_event(event: Event, *args) -> list:
    out = []
    for handler in custom_event_handlers[event]:
        try:
            out.append(await handler(*args))
        except Exception as e:
            print(f'\nerror has occurred while triggering a custom event on a mod, details:\n'
                  f'event: {event}\n'
                  f'error: {type(e)}\n'
                  f'reason: {e}\n'
                  f'stack trace:')
            traceback.print_exc()
    return out


def event_handler(event: Event):
    def _register(func):
        custom_event_handlers[event].append(func)
        return AsyncEventWrapper(func, type=event)

    return _register


class AsyncEventWrapper:
    def __init__(self, func, type: Event):
        self.func = func
        self.type = type

    def unregister(self):
        events = custom_event_handlers[self.type]
        if self.func in events:
            events.remove(self.func)

    async def __call__(self, *args, **kwargs):
        await self.func(args, **kwargs)



from datetime import datetime

from .baseapi import Api
from .. import util


class StreamInfoApi(Api):
    def __init__(self, client_id: str, user: str):
        super().__init__(client_id, user)

        self.viewer_count: int = 0
        self.title: str = ''
        self.game_id: int = 0
        self.community_ids: frozenset = frozenset()
        self.started_at: datetime = datetime.min
        self.user_id: int = 0
        self.tag_ids: frozenset = frozenset()

    async def update(self, log=False):
        """
        requests the updated stream info from twitch, called every X seconds (default 60)

        calls `self.on_successful_update` when it updates without errors

        calls `self.on_failed_update` when the update fails, (usually due to key errors)

        :param log: should errors be logged?
        """
        data = await util.get_stream_data(self.user, self.headers)
        try:
            self.viewer_count = data['viewer_count']
            self.title = data['title']
            self.game_id = data['game_id']
            self.community_ids = frozenset(data['community_ids'])
            #                                                      2018-05-17T16:47:46Z
            self.started_at = datetime.strptime(data['started_at'], '%Y-%m-%dT%H:%M:%SZ')
            self.user_id = data['user_id']
            self.tag_ids = frozenset(data['tag_ids'])

            await self.on_successful_update()
        except KeyError as e:
            if log:
                print(f'[STREAM INFO API] failed to update stream data for {self.user}')
                print(f'[STREAM INFO API] key error occurred while trying to access key: {e}')

            await self.on_failed_update()



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md') as f:
    long_description = f.read()

setup(
    name='PythonTwitchBotFramework',
    version='1.11.11',
    python_requires='>=3.6',
    packages=find_packages(),
    url='https://github.com/sharkbound/PythonTwitchBotFramework',
    license='MIT',
    author='sharkbound',
    description='asynchronous twitch-bot framework made in pure python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    scripts=[
        'util/command_console.py',
    ]
)



from twitchbot import (
    Command,
    perms,
    Message,
    cfg,
    InvalidArgumentsError)

WHISPER = True

PREFIX = cfg.prefix
PERMISSION = 'manage_permissions'


@Command('addperm', permission=PERMISSION, syntax='<group> <permission>', help='adds a permission a group')
async def cmd_add_perm(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_add_perm)

    group, perm = args
    if perms.add_permission(msg.channel_name, group, perm):
        await msg.reply(whisper=WHISPER, msg=f'added permission "{perm}" to "{group}"')
    else:
        await msg.reply(whisper=WHISPER, msg=f'no group found by the "{group}"')


@Command('delperm', permission=PERMISSION, syntax='<group> <permission>', help='removes a permission from a group')
async def cmd_del_perm(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_del_perm)

    group, perm = args
    if perms.delete_permission(msg.channel_name, group, perm):
        await msg.reply(whisper=WHISPER, msg=f'deleted permission "{perm}" from "{group}"')
    else:
        await msg.reply(whisper=WHISPER, msg=f'no group found by the "{group}"')


@Command('addgroup', permission=PERMISSION, syntax='<group>', help='adds a permission group')
async def cmd_add_group(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_add_group)

    group = args[0]
    if perms.add_group(msg.channel_name, group):
        await msg.reply(whisper=WHISPER, msg=f'added permission group "{group}"')
    else:
        await msg.reply(whisper=WHISPER, msg=f'permission group "{group}" already exist')


@Command('delgroup', permission=PERMISSION, syntax='<group>', help='removes a permission group')
async def cmd_del_group(msg: Message, *args):
    if not args:
        raise InvalidArgumentsError(reason='missing required argument', cmd=cmd_del_group)

    group = args[0]
    if perms.delete_group(msg.channel_name, group):
        await msg.reply(whisper=WHISPER, msg=f'deleted permission group "{group}"')
    else:
        await msg.reply(whisper=WHISPER, msg=f'no group found by the "{group}"')


@Command('reloadperms', permission=PERMISSION, help='reloads permissions')
async def cmd_reload_perms(msg: Message, *args):
    if perms.reload_permissions(channel=msg.channel_name):
        await msg.reply(whisper=WHISPER, msg='reloaded permissions')
    else:
        await msg.reply(whisper=WHISPER, msg='failed to reload permissions')


@Command('addmember', permission=PERMISSION, syntax='<group> <member>', help='adds a member to a permission group')
async def cmd_add_member(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_add_member)

    group, member = args
    member = member.lstrip('@')

    if perms.add_member(msg.channel_name, group, member):
        await msg.reply(whisper=WHISPER, msg=f'added "{member}" to "{group}"')
    else:
        await msg.reply(whisper=WHISPER, msg=f'failed to add member, group does not exist')


@Command('delmember', permission=PERMISSION, syntax='<group> <member>', help='removes a member from a permission group')
async def cmd_del_member(msg: Message, *args):
    if len(args) != 2:
        raise InvalidArgumentsError(reason='missing required arguments', cmd=cmd_del_member)

    group, member = args
    member = member.lstrip('@')

    if perms.delete_member(msg.channel_name, group, member):
        await msg.reply(whisper=WHISPER, msg=f'removed "{member}" from "{group}"')
        return

    g = perms.get_group(msg.channel_name, group)
    await msg.reply(whisper=WHISPER,
                    msg='failed to remove member, group does not exist' if not g else f'"{member}" is not in that group')



red = 'Red'
blue = 'Blue'
green = 'Green'
spring_green = 'SpringGreen'
hot_pink = 'HotPink'
blue_violet = 'BlueViolet'
cadet_blue = 'CadetBlue'
chocolate = 'Chocolate'
coral = 'Coral'
dodger_blue = 'DodgerBlue'
firebrick = 'Firebrick'
golden_rod = 'GoldenRod'
orange_red = 'OrangeRed'
yellow_green = 'YellowGreen'
sea_green = 'SeaGreen'


from typing import List
from .follower import Follower

__all__ = ['UserFollowers']


class UserFollowers:
    __slots__ = ['followers', 'id', 'name', 'follower_count']

    def __init__(self, follower_count: int, following: str, following_id: int, name: str, id: int,
                 followers: List[dict]):
        self.followers: List[Follower] = [
            Follower(following=following,
                     following_id=int(following_id),
                     id=int(json['from_id']),
                     name=json['from_name'])
            for json in followers
            if 'from_id' in json and 'from_name' in json
        ]
        self.id: int = id
        self.name: str = name
        self.follower_count = follower_count

    def __iter__(self):
        yield from self.followers



from typing import Dict
from dataclasses import dataclass, field
from .util import get_url


@dataclass(frozen=True)
class Emote:
    id: int
    code: str
    set: int = field(default=0, repr=False)


GLOBAL_EMOTE_API = 'https://api.twitchemotes.com/api/v4/channels/0'
emotes: Dict[str, Emote] = {}


async def update_global_emotes():
    _, data = await get_url(GLOBAL_EMOTE_API)
    emotes.clear()

    for emote in data['emotes']:
        emotes[emote['code']] = Emote(int(emote['id']), emote['code'], emote['emoticon_set'])



from twitchbot import (
    Message,
    Command,
    CommandContext,
    InvalidArgumentsError,
    is_mod_disabled,
    mod_exists,
    enable_mod,
    disable_mod,
    mods,
    reload_mod)

MOD_MANAGE_PERMISSION = 'manage_mods'


@Command('enablemod', context=CommandContext.CHANNEL, permission=MOD_MANAGE_PERMISSION,
         syntax='<mod_name>', help='enables a mod for the current channel')
async def cmd_enable_mod(msg: Message, *args):
    if len(args) != 1:
        raise InvalidArgumentsError(reason=f'expected one argument, got {len(args)}', cmd=cmd_enable_mod)

    mod = args[0]
    if not mod_exists(mod):
        raise InvalidArgumentsError(reason=f'could not find mod "{mod}"')

    if not is_mod_disabled(msg.channel_name, mod):
        await msg.reply(f'mod "{mod}" is already enabled')
        return

    enable_mod(msg.channel_name, mod)
    await msg.reply(f'mod "{mod}" has been enabled for this channel')


@Command('disablemod', context=CommandContext.CHANNEL, permission=MOD_MANAGE_PERMISSION,
         syntax='<mod_name>', help='disables a mod for the current channel')
async def cmd_disable_mod(msg: Message, *args):
    if len(args) != 1:
        raise InvalidArgumentsError(reason=f'expected one argument, got {len(args)}', cmd=cmd_disable_mod)

    mod = args[0]
    if not mod_exists(mod):
        raise InvalidArgumentsError(reason=f'could not find mod "{mod}"', cmd=cmd_disable_mod)

    if is_mod_disabled(msg.channel_name, mod):
        await msg.reply(f'mod "{mod}" is already disabled')
        return

    disable_mod(msg.channel_name, mod)
    await msg.reply(f'mod "{mod}" has been disabled for this channel')


@Command('disabledmods', context=CommandContext.CHANNEL, help='lists disabled mods for the current channel')
async def cmd_disabled_mods(msg: Message, *args):
    await msg.reply(f'disabled mods for "{msg.channel_name}": ' + ', '.join(
        mod for mod in mods if is_mod_disabled(msg.channel_name, mod)))


@Command('reloadmod', context=CommandContext.CHANNEL, syntax='<modname(case sensitive) or all>',
         permission=MOD_MANAGE_PERMISSION,
         help='reloads a specific module (case sensitive) from disk')
async def cmd_reload_mod(msg: Message, *args):
    if len(args) != 1:
        raise InvalidArgumentsError(f'missing required argument: modname', cmd=cmd_reload_mod)
    mod = args[0]

    if mod != 'all' and not mod_exists(mod):
        raise InvalidArgumentsError(f'could not find any mod by name: {mod}', cmd=cmd_reload_mod)

    if mod == 'all':
        for mod in tuple(mods):
            if not reload_mod(mod):
                await msg.reply(f'failed to reload mod "{mod}"')
                return
        await msg.reply('reloaded all mods')
        return

    if reload_mod(mod):
        await msg.reply(f'reloaded mod "{mod}"')
    else:
        await msg.reply(f'failed to reload mod "{mod}"')



from typing import Dict, Tuple
from .models import Balance, CurrencyName
from .session import session

__all__ = [
    'get_balance',
    'set_balance',
    'get_currency_name',
    'add_balance',
    'add_balance_to_all',
    'get_balance_from_msg',
    'set_currency_name',
    'subtract_balance',
    'subtract_balance_from_all',
]

currency_name_cache: Dict[str, CurrencyName] = {}


def add_balance_to_all(channel: str, value: int):
    session.query(Balance) \
        .filter(Balance.channel == channel) \
        .update({Balance.balance: Balance.balance + value})
    session.commit()


def subtract_balance_from_all(channel: str, value: int):
    session.query(Balance) \
        .filter(Balance.channel == channel) \
        .update({Balance.balance: Balance.balance - value})
    session.commit()


def set_balance(channel: str, user: str, value: int):
    """sets a users balance"""
    get_balance(channel, user.lower()).balance = max(0, value)
    session.commit()


def add_balance(channel: str, user: str, value: int, commit=True):
    """adds balance to a user in the specified channel"""

    get_balance(channel, user.lower()).balance += value
    if commit:
        session.commit()


def subtract_balance(channel: str, user: str, value: int):
    """subtracts balance to a user in the specified channel"""

    get_balance(channel, user.lower()).balance -= value
    session.commit()


def get_balance(channel: str, user: str) -> Balance:
    """gets the balance of the user for the specified channel"""

    user = user.lower()
    bal = session.query(Balance).filter(Balance.channel == channel, Balance.user == user).one_or_none()

    if bal is None:
        bal = Balance.create(channel, user)
        session.add(bal)
        session.commit()

    return bal


def get_balance_from_msg(msg) -> Balance:
    """gets the balance for the user from the channel the msg came from"""

    return get_balance(msg.channel_name, msg.author.lower())


def get_currency_name(channel: str) -> CurrencyName:
    """returns a CurrencyName object for the channel specifed, returns None if it doesnt exist"""

    if channel in currency_name_cache:
        return currency_name_cache[channel]

    currency = session.query(CurrencyName).filter(CurrencyName.channel == channel).one_or_none()
    if currency is None:
        currency = CurrencyName.create(channel, 'points')
        session.add(currency)
        session.commit()

    currency_name_cache[channel] = currency
    return currency


def set_currency_name(channel: str, new_name: str) -> bool:
    """sets a channels currency name, return if it was successful"""

    if not new_name:
        return False

    currency = get_currency_name(channel)

    if currency is None:
        session.add(CurrencyName.create(channel, new_name))
    else:
        currency.name = new_name

    currency_name_cache[channel] = currency
    session.commit()
    return True
